# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from oslo_utils import timeutils
import webob

from cinder.api.v3 import types
from cinder.api.v3.views import types as views_types
from cinder import context
from cinder import exception
from cinder.policies import type_extra_specs as extra_specs_policy
from cinder.policies import volume_type as type_policy
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.volume import volume_types


def fake_volume_type(id):
    specs = {
        "key1": "value1",
        "key2": "value2",
        "key3": "value3",
        "key4": "value4",
        "key5": "value5"
    }
    return dict(
        id=id,
        name='vol_type_%s' % id,
        description='vol_type_desc_%s' % id,
        extra_specs=specs,
    )


def return_volume_types_get_all_types(context, filters=None, marker=None,
                                      limit=None, sort_keys=None,
                                      sort_dirs=None, offset=None,
                                      list_result=False):
    result = dict(vol_type_1=fake_volume_type(1),
                  vol_type_2=fake_volume_type(2),
                  vol_type_3=fake_volume_type(3)
                  )
    if list_result:
        return list(result.values())
    return result


def return_empty_volume_types_get_all_types(context, filters=None, marker=None,
                                            limit=None, sort_keys=None,
                                            sort_dirs=None, offset=None,
                                            list_result=False):
    if list_result:
        return []
    return {}


def return_volume_types_get_volume_type(context, id):
    if id == fake.WILL_NOT_BE_FOUND_ID:
        raise exception.VolumeTypeNotFound(volume_type_id=id)
    return fake_volume_type(id)


def return_volume_types_get_default(context):
    return fake_volume_type(1)


@ddt.ddt
class VolumeTypesApiTest(test.TestCase):

    def _create_volume_type(self, volume_type_name, extra_specs=None,
                            is_public=True, projects=None):
        return volume_types.create(self.ctxt, volume_type_name, extra_specs,
                                   is_public, projects).get('id')

    def setUp(self):
        super(VolumeTypesApiTest, self).setUp()
        self.controller = types.VolumeTypesController()
        self.ctxt = context.RequestContext(user_id=fake.USER_ID,
                                           project_id=fake.PROJECT_ID,
                                           is_admin=True)
        self.mock_authorize = self.patch(
            'cinder.context.RequestContext.authorize')
        # since __DEFAULT__ type always exists, total number of volume types
        # is total_types_created + 1. In this case it's 4
        self.type_id1 = self._create_volume_type('volume_type1',
                                                 {'key1': 'value1'})
        self.type_id2 = self._create_volume_type('volume_type2',
                                                 {'key2': 'value2'})
        self.type_id3 = self._create_volume_type('volume_type3',
                                                 {'key3': 'value3'}, False,
                                                 [fake.PROJECT_ID])
        self.default_type = volume_types.get_default_volume_type()['id']
        self.vol_type = volume_types.get_by_name_or_id(
            context.get_admin_context(), '__DEFAULT__')['id']

    def test_volume_types_index(self):
        self.mock_object(volume_types, 'get_all_types',
                         return_volume_types_get_all_types)

        req = fakes.HTTPRequest.blank('/v3/%s/types' % fake.PROJECT_ID,
                                      use_admin_context=True)
        res_dict = self.controller.index(req)

        self.assertEqual(3, len(res_dict['volume_types']))

        expected_names = ['vol_type_1', 'vol_type_2', 'vol_type_3']
        actual_names = map(lambda e: e['name'], res_dict['volume_types'])
        self.assertEqual(set(expected_names), set(actual_names))
        for entry in res_dict['volume_types']:
            self.assertEqual('value1', entry['extra_specs']['key1'])
        self.mock_authorize.assert_any_call(type_policy.GET_ALL_POLICY)

    def test_volume_types_index_no_data(self):
        self.mock_object(volume_types, 'get_all_types',
                         return_empty_volume_types_get_all_types)

        req = fakes.HTTPRequest.blank('/v3/%s/types' % fake.PROJECT_ID)
        res_dict = self.controller.index(req)

        self.assertEqual(0, len(res_dict['volume_types']))

    def test_volume_types_index_with_limit(self):
        req = fakes.HTTPRequest.blank('/v3/%s/types?limit=1' % fake.PROJECT_ID)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(1, len(res['volume_types']))
        self.assertEqual(self.type_id3, res['volume_types'][0]['id'])

        expect_next_link = ('http://localhost/v3/%s/types?limit=1'
                            '&marker=%s' %
                            (fake.PROJECT_ID, res['volume_types'][0]['id']))
        self.assertEqual(expect_next_link, res['volume_type_links'][0]['href'])

    def test_volume_types_index_with_offset(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types?offset=1' % fake.PROJECT_ID)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(3, len(res['volume_types']))

    def test_volume_types_index_with_offset_out_of_range(self):
        url = '/v3/%s/types?offset=424366766556787' % fake.PROJECT_ID
        req = fakes.HTTPRequest.blank(url)
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index, req)

    def test_volume_types_index_with_limit_and_offset(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types?limit=2&offset=1' % fake.PROJECT_ID)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(2, len(res['volume_types']))
        self.assertEqual(self.type_id2, res['volume_types'][0]['id'])
        self.assertEqual(self.type_id1, res['volume_types'][1]['id'])

    def test_volume_types_index_with_limit_and_marker(self):
        req = fakes.HTTPRequest.blank('/v3/%s/types?limit=1'
                                      '&marker=%s' %
                                      (fake.PROJECT_ID,
                                       self.type_id2))
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(1, len(res['volume_types']))
        self.assertEqual(self.type_id1, res['volume_types'][0]['id'])

    def test_volume_types_index_with_valid_filter(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types?is_public=True' % fake.PROJECT_ID)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        self.assertEqual(4, len(res['volume_types']))
        self.assertEqual(self.type_id3, res['volume_types'][0]['id'])
        self.assertEqual(self.type_id2, res['volume_types'][1]['id'])
        self.assertEqual(self.type_id1, res['volume_types'][2]['id'])

    def test_volume_types_index_with_invalid_filter(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types?id=%s' % (fake.PROJECT_ID, self.type_id1))
        req.environ['cinder.context'] = context.RequestContext(
            user_id=fake.USER_ID, project_id=fake.PROJECT_ID, is_admin=False)
        res = self.controller.index(req)

        self.assertEqual(4, len(res['volume_types']))

    def test_volume_types_index_with_sort_keys(self):
        req = fakes.HTTPRequest.blank('/v3/%s/types?sort=id' % fake.PROJECT_ID)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)
        expect_result = [self.default_type, self.type_id1,
                         self.type_id2, self.type_id3]
        expect_result.sort(reverse=True)

        self.assertEqual(4, len(res['volume_types']))
        self.assertEqual(expect_result[0], res['volume_types'][0]['id'])
        self.assertEqual(expect_result[1], res['volume_types'][1]['id'])
        self.assertEqual(expect_result[2], res['volume_types'][2]['id'])
        self.assertEqual(expect_result[3], res['volume_types'][3]['id'])

    def test_volume_types_index_with_sort_and_limit(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types?sort=id&limit=2' % fake.PROJECT_ID)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)

        expect_result = [self.default_type, self.type_id1,
                         self.type_id2, self.type_id3]
        expect_result.sort(reverse=True)

        self.assertEqual(2, len(res['volume_types']))
        self.assertEqual(expect_result[0], res['volume_types'][0]['id'])
        self.assertEqual(expect_result[1], res['volume_types'][1]['id'])

    def test_volume_types_index_with_sort_keys_and_sort_dirs(self):
        req = fakes.HTTPRequest.blank(
            '/v3/%s/types?sort=id:asc' % fake.PROJECT_ID)
        req.environ['cinder.context'] = self.ctxt
        res = self.controller.index(req)
        expect_result = [self.default_type, self.type_id1,
                         self.type_id2, self.type_id3]
        expect_result.sort()

        self.assertEqual(4, len(res['volume_types']))
        self.assertEqual(expect_result[0], res['volume_types'][0]['id'])
        self.assertEqual(expect_result[1], res['volume_types'][1]['id'])
        self.assertEqual(expect_result[2], res['volume_types'][2]['id'])
        self.assertEqual(expect_result[3], res['volume_types'][3]['id'])

    def test_volume_types_show(self):
        self.mock_object(volume_types, 'get_volume_type',
                         return_volume_types_get_volume_type)

        type_id = fake.VOLUME_TYPE_ID
        req = fakes.HTTPRequest.blank('/v3/%s/types/' % fake.PROJECT_ID
                                      + type_id)
        res_dict = self.controller.show(req, type_id)

        self.assertEqual(1, len(res_dict))
        self.assertEqual(type_id, res_dict['volume_type']['id'])
        type_name = 'vol_type_' + type_id
        self.assertEqual(type_name, res_dict['volume_type']['name'])
        self.mock_authorize.assert_any_call(
            type_policy.GET_POLICY, target_obj=mock.ANY)

    def test_volume_types_show_not_found(self):
        self.mock_object(volume_types, 'get_volume_type',
                         return_volume_types_get_volume_type)

        req = fakes.HTTPRequest.blank('/v3/%s/types/%s' %
                                      (fake.PROJECT_ID,
                                       fake.WILL_NOT_BE_FOUND_ID))
        self.assertRaises(exception.VolumeTypeNotFound, self.controller.show,
                          req, fake.WILL_NOT_BE_FOUND_ID)

    def test_get_default(self):
        self.mock_object(volume_types, 'get_default_volume_type',
                         return_volume_types_get_default)
        req = fakes.HTTPRequest.blank('/v3/%s/types/default' % fake.PROJECT_ID)
        req.method = 'GET'
        res_dict = self.controller.show(req, 'default')
        self.assertEqual(1, len(res_dict))
        self.assertEqual('vol_type_1', res_dict['volume_type']['name'])
        self.assertEqual('vol_type_desc_1',
                         res_dict['volume_type']['description'])

    def test_get_default_not_found(self):
        self.mock_object(volume_types, 'get_default_volume_type',
                         return_value={})
        req = fakes.HTTPRequest.blank('/v3/%s/types/default' % fake.PROJECT_ID)
        req.method = 'GET'

        self.assertRaises(exception.VolumeTypeNotFound,
                          self.controller.show, req, 'default')

    @ddt.data(
        {
            'extra_spec_policy': False,
            'read_sensitive_policy': False,
            'qos_policy': False,
        },
        {
            'extra_spec_policy': True,
            'read_sensitive_policy': False,
            'qos_policy': False,
        },
        {
            'extra_spec_policy': True,
            'read_sensitive_policy': True,
            'qos_policy': False,
        },
        {
            'extra_spec_policy': False,
            'read_sensitive_policy': False,
            'qos_policy': True,
        },
        {
            'extra_spec_policy': True,
            'read_sensitive_policy': True,
            'qos_policy': True,
        },
    )
    @ddt.unpack
    def test_view_builder_show(self,
                               extra_spec_policy,
                               read_sensitive_policy,
                               qos_policy):
        # This function returns the authorization result supplied by the
        # DDT data for the associated policy.
        def authorize(policy, fatal):
            policy_data = {
                type_policy.EXTRA_SPEC_POLICY: extra_spec_policy,
                extra_specs_policy.READ_SENSITIVE_POLICY: (
                    read_sensitive_policy),
                type_policy.QOS_POLICY: qos_policy,
            }
            return policy_data[policy]

        view_builder = views_types.ViewBuilder()
        now = timeutils.utcnow().isoformat()
        raw_volume_type = dict(
            name='new_type',
            description='new_type_desc',
            qos_specs_id='new_id',
            is_public=True,
            deleted=False,
            created_at=now,
            updated_at=now,
            extra_specs={'multiattach': True, 'sensitive': 'secret'},
            deleted_at=None,
            id=42,
        )

        request = fakes.HTTPRequest.blank("/v3")
        with mock.patch('cinder.context.RequestContext.authorize',
                        side_effect=authorize):
            output = view_builder.show(request, raw_volume_type)

        self.assertIn('volume_type', output)
        expected_volume_type = dict(
            name='new_type',
            description='new_type_desc',
            is_public=True,
            id=42,
        )
        if extra_spec_policy:
            expected_volume_type['extra_specs'] = {'multiattach': True}
            if read_sensitive_policy:
                expected_volume_type['extra_specs']['sensitive'] = 'secret'
        if qos_policy:
            expected_volume_type['qos_specs_id'] = 'new_id'

        self.assertDictEqual(expected_volume_type, output['volume_type'])

    @ddt.data(False, True)
    def test_view_builder_list(self, is_admin):
        view_builder = views_types.ViewBuilder()
        self.mock_authorize.return_value = is_admin
        now = timeutils.utcnow().isoformat()
        raw_volume_types = []
        for i in range(0, 10):
            raw_volume_types.append(
                dict(
                    name='new_type',
                    description='new_type_desc',
                    qos_specs_id='new_id',
                    is_public=True,
                    deleted=False,
                    created_at=now,
                    updated_at=now,
                    extra_specs={'multiattach': True, 'sensitive': 'secret'},
                    deleted_at=None,
                    id=42 + i
                )
            )

        request = fakes.HTTPRequest.blank("/v3")
        output = view_builder.index(request, raw_volume_types)

        self.assertIn('volume_types', output)
        for i in range(0, 10):
            expected_volume_type = dict(
                name='new_type',
                description='new_type_desc',
                is_public=True,
                id=42 + i
            )
            if is_admin:
                expected_volume_type['qos_specs_id'] = 'new_id'
                expected_volume_type['extra_specs'] = {'multiattach': True,
                                                       'sensitive': 'secret'}
            self.assertDictEqual(expected_volume_type,
                                 output['volume_types'][i])



# Copyright 2017 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Volume driver for Inspur AS13000
"""

import ipaddress
import json
import random
import re
import time

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units
import requests

from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
from cinder.volume.drivers.san import san
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)

inspur_as13000_opts = [
    cfg.ListOpt(
        'as13000_ipsan_pools',
        default=['Pool0'],
        help='The Storage Pools Cinder should use, a comma separated list.'),
    cfg.IntOpt(
        'as13000_token_available_time',
        default=3300,
        min=600, max=3600,
        help='The effective time of token validity in seconds.'),
    cfg.StrOpt(
        'as13000_meta_pool',
        help='The pool which is used as a meta pool when creating a volume, '
             'and it should be a replication pool at present. '
             'If not set, the driver will choose a replication pool '
             'from the value of as13000_ipsan_pools.'),
]

CONF = cfg.CONF
CONF.register_opts(inspur_as13000_opts)


class RestAPIExecutor(object):
    def __init__(self, hostname, port, username, password):
        self._username = username
        self._password = password
        self._token = None
        self._baseurl = 'http://%s:%s/rest' % (hostname, port)

    def login(self):
        """Login the AS13000 and store the token."""
        self._token = self._login()
        LOG.debug('Login the AS13000.')

    def _login(self):
        """Do request to login the AS13000 and get the token."""
        method = 'security/token'
        params = {'name': self._username, 'password': self._password}
        token = self.send_rest_api(method=method, params=params,
                                   request_type='post').get('token')
        return token

    @utils.retry(exception.VolumeDriverException, interval=1, retries=3)
    def send_rest_api(self, method, params=None, request_type='post'):
        try:
            return self.send_api(method, params, request_type)
        except exception.VolumeDriverException:
            self.login()
            raise

    @staticmethod
    @volume_utils.trace_method
    def do_request(cmd, url, header, data):
        """Send request to the storage and handle the response."""
        if cmd in ['post', 'get', 'put', 'delete']:
            req = getattr(requests, cmd)(url, data=data, headers=header)
        else:
            msg = (_('Unsupported cmd: %s.') % cmd)
            raise exception.VolumeBackendAPIException(msg)

        response = req.json()
        code = req.status_code
        LOG.debug('CODE: %(code)s, RESPONSE: %(response)s.',
                  {'code': code, 'response': response})

        if code != 200:
            msg = (_('Code: %(code)s, URL: %(url)s, Message: %(msg)s.')
                   % {'code': req.status_code,
                      'url': req.url,
                      'msg': req.text})
            LOG.error(msg)
            raise exception.VolumeDriverException(msg)

        return response

    @volume_utils.trace
    def send_api(self, method, params=None, request_type='post'):
        if params:
            params = json.dumps(params)

        url = '%s/%s' % (self._baseurl, method)

        # header is not needed when the driver login the backend
        if method == 'security/token':
            if request_type == 'delete':
                header = {'X-Auth-Token': self._token}
            else:
                header = None
        else:
            if not self._token:
                self.login()
            header = {'X-Auth-Token': self._token}

        response = self.do_request(request_type, url, header, params)

        try:
            code = response.get('code')
            if code == 0:
                if request_type == 'get':
                    data = response.get('data')
                else:
                    if method == 'security/token':
                        data = response.get('data')
                    else:
                        data = response.get('message')
                        data = str(data).lower()
                        if hasattr(data, 'success'):
                            return
            elif code == 301:
                msg = _('Token is expired.')
                LOG.error(msg)
                raise exception.VolumeDriverException(msg)
            else:
                message = response.get('message')
                msg = (_('Unexpected RestAPI response: %(code)d %(msg)s.') % {
                       'code': code, 'msg': message})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(msg)
        except ValueError:
            msg = _("Deal with response failed.")
            raise exception.VolumeDriverException(msg)

        return data


@interface.volumedriver
class AS13000Driver(san.SanISCSIDriver):
    """Driver for Inspur AS13000 storage.

    .. code-block:: none

      Version history:
          1.0.0 - Initial driver

    """

    VENDOR = 'INSPUR'
    VERSION = '1.0.0'
    PROTOCOL = 'iSCSI'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = 'Inspur_CI'

    def __init__(self, *args, **kwargs):
        super(AS13000Driver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(inspur_as13000_opts)
        self.hostname = self.configuration.san_ip
        self.port = self.configuration.safe_get('san_api_port') or 8088
        self.username = self.configuration.san_login
        self.password = self.configuration.san_password
        self.token_available_time = (self.configuration.
                                     as13000_token_available_time)
        self.pools = self.configuration.as13000_ipsan_pools
        self.meta_pool = self.configuration.as13000_meta_pool
        self.pools_info = {}
        self.nodes = []
        self._token_time = 0
        # get the RestAPIExecutor
        self._rest = RestAPIExecutor(self.hostname,
                                     self.port,
                                     self.username,
                                     self.password)

    @staticmethod
    def get_driver_options():
        return inspur_as13000_opts

    @volume_utils.trace
    def do_setup(self, context):
        # get tokens for the driver
        self._rest.login()
        self._token_time = time.time()

        # get available nodes in the backend
        for node in self._get_cluster_status():
            if node.get('healthStatus') == 1 and node.get('ip'):
                self.nodes.append(node)

        # collect pools info
        meta_pools = [self.meta_pool] if self.meta_pool else []
        self.pools_info = self._get_pools_info(self.pools + meta_pools)

        # setup the meta pool if it is not setted
        if not self.meta_pool:
            for pool_info in self.pools_info.values():
                if pool_info['type'] in (1, '1'):
                    self.meta_pool = pool_info['name']
                    break

        self._check_pools()

        self._check_meta_pool()

    @volume_utils.trace
    def check_for_setup_error(self):
        """Do check to make sure service is available."""
        # check the required flags in conf
        required_flags = ['san_ip', 'san_login', 'san_password',
                          'as13000_ipsan_pools']
        for flag in required_flags:
            value = self.configuration.safe_get(flag)
            if not value:
                msg = (_('Required flag %s is not set.') % flag)
                LOG.error(msg)
                raise exception.InvalidConfigurationValue(option=flag,
                                                          value=value)

        # make sure at least one node can
        if not self.nodes:
            msg = _('No healthy nodes are available!')
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

    def _check_pools(self):
        """Check the pool in conf exist in the AS13000."""
        if not set(self.pools).issubset(self.pools_info):
            pools = set(self.pools) - set(self.pools_info)
            msg = _('Pools %s do not exist.') % pools
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

    def _check_meta_pool(self):
        """Check whether the meta pool is valid."""
        if not self.meta_pool:
            msg = _('Meta pool is not set.')
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        if self.meta_pool not in self.pools_info:
            msg = _('Meta pool %s does not exist.') % self.meta_pool
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        if self.pools_info[self.meta_pool]['type'] not in (1, '1'):
            msg = _('Meta pool %s is not a replication pool.') % self.meta_pool
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

    @volume_utils.trace
    def create_volume(self, volume):
        """Create volume in the backend."""
        pool = volume_utils.extract_host(volume.host, level='pool')
        size = volume.size * units.Ki
        name = self._trans_name_down(volume.name)

        method = 'block/lvm'
        request_type = "post"
        params = {
            "name": name,
            "capacity": size,
            "dataPool": pool,
            "dataPoolType": self.pools_info[pool]['type'],
            "metaPool": self.meta_pool
        }
        self._rest.send_rest_api(method=method, params=params,
                                 request_type=request_type)

    @volume_utils.trace
    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a new volume base on a specific snapshot."""
        if snapshot.volume_size > volume.size:
            msg = (_("create_volume_from_snapshot: snapshot %(snapshot_name)s "
                     "size is %(snapshot_size)dGB and doesn't fit in target "
                     "volume %(volume_name)s of size %(volume_size)dGB.") %
                   {'snapshot_name': snapshot.name,
                    'snapshot_size': snapshot.volume_size,
                    'volume_name': volume.name,
                    'volume_size': volume.size})
            LOG.error(msg)
            raise exception.InvalidInput(message=msg)
        src_vol_name = self._trans_name_down(snapshot.volume_name)
        source_vol = snapshot.volume
        src_pool = volume_utils.extract_host(source_vol['host'],
                                             level='pool')
        dest_name = self._trans_name_down(volume.name)
        dest_pool = volume_utils.extract_host(volume.host, level='pool')
        snap_name = self._trans_name_down(snapshot.name)

        # lock the snapshot before clone from it
        self._snapshot_lock_op('lock', src_vol_name, snap_name, src_pool)

        # do clone from snap to a volume
        method = 'snapshot/volume/cloneLvm'
        request_type = 'post'
        params = {'originalLvm': src_vol_name,
                  'originalPool': src_pool,
                  'originalSnap': snap_name,
                  'name': dest_name,
                  'pool': dest_pool}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

        # do filling the cloned volume
        self._filling_volume(dest_name, dest_pool)

        # wait until the cloned volume has been filled
        self._wait_volume_filled(dest_name, dest_pool)

        # unlock the original snapshot
        self._snapshot_lock_op('unlock', src_vol_name, snap_name, src_pool)

        if volume.size > snapshot.volume_size:
            self.extend_volume(volume, volume.size)

    @volume_utils.trace
    def create_cloned_volume(self, volume, src_vref):
        """Clone a volume."""
        if src_vref.size > volume.size:
            msg = (_("create_cloned_volume: source volume %(src_vol)s "
                     "size is %(src_size)dGB and doesn't fit in target "
                     "volume %(tgt_vol)s of size %(tgt_size)dGB.") %
                   {'src_vol': src_vref.name,
                    'src_size': src_vref.size,
                    'tgt_vol': volume.name,
                    'tgt_size': volume.size})
            LOG.error(msg)
            raise exception.InvalidInput(message=msg)
        dest_pool = volume_utils.extract_host(volume.host, level='pool')
        dest_vol_name = self._trans_name_down(volume.name)
        src_pool = volume_utils.extract_host(src_vref.host, level='pool')
        src_vol_name = self._trans_name_down(src_vref.name)

        method = 'block/lvm/clone'
        request_type = 'post'
        params = {'srcVolumeName': src_vol_name,
                  'srcPoolName': src_pool,
                  'destVolumeName': dest_vol_name,
                  'destPoolName': dest_pool}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

        if volume.size > src_vref.size:
            self.extend_volume(volume, volume.size)

    @volume_utils.trace
    def extend_volume(self, volume, new_size):
        """Extend volume to new size."""
        name = self._trans_name_down(volume.name)
        if not self._check_volume(volume):
            msg = _('Extend Volume Failed: Volume %s does not exist.') % name
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        size = new_size * units.Ki
        pool = volume_utils.extract_host(volume.host, level='pool')

        method = 'block/lvm'
        request_type = 'put'
        params = {'pool': pool,
                  'name': name,
                  'newCapacity': size}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @volume_utils.trace
    def delete_volume(self, volume):
        """Delete volume from AS13000."""
        name = self._trans_name_down(volume.name)
        if not self._check_volume(volume):
            # if volume is not exist in backend, the driver will do
            # nothing but log it
            LOG.info('Tried to delete non-existent volume %(name)s.',
                     {'name': name})
            return

        pool = volume_utils.extract_host(volume.host, level='pool')

        method = 'block/lvm?pool=%s&lvm=%s' % (pool, name)
        request_type = 'delete'
        self._rest.send_rest_api(method=method, request_type=request_type)

    @volume_utils.trace
    def create_snapshot(self, snapshot):
        """Create snapshot of volume in backend.

        The snapshot type of AS13000 is copy-on-write.
        """
        source_volume = snapshot.volume
        volume_name = self._trans_name_down(source_volume.name)
        if not self._check_volume(source_volume):
            msg = (_('create_snapshot: Source_volume %s does not exist.')
                   % volume_name)
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        pool = volume_utils.extract_host(source_volume.host, level='pool')
        snapshot_name = self._trans_name_down(snapshot.name)

        method = 'snapshot/volume'
        request_type = 'post'
        params = {'snapName': snapshot_name,
                  'volumeName': volume_name,
                  'poolName': pool,
                  'snapType': 'r'}
        self._rest.send_rest_api(method=method, params=params,
                                 request_type=request_type)

    @volume_utils.trace
    def delete_snapshot(self, snapshot):
        """Delete snapshot of volume."""
        source_volume = snapshot.volume
        volume_name = self._trans_name_down(source_volume.name)
        if self._check_volume(source_volume) is False:
            msg = (_('delete_snapshot: Source_volume %s does not exist.')
                   % volume_name)
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        pool = volume_utils.extract_host(source_volume.host, level='pool')
        snapshot_name = self._trans_name_down(snapshot.name)

        method = ('snapshot/volume?snapName=%s&volumeName=%s&poolName=%s'
                  % (snapshot_name, volume_name, pool))
        request_type = 'delete'
        self._rest.send_rest_api(method=method, request_type=request_type)

    @volume_utils.trace
    def _update_volume_stats(self):
        """Update the backend stats including driver info and pools info."""

        # As _update_volume_stats runs periodically,
        # so we can do a check and refresh the token each time it runs.
        time_difference = time.time() - self._token_time
        if time_difference > self.token_available_time:
            self._rest.login()
            self._token_time = time.time()
            LOG.debug('Token of the Driver has been refreshed.')

        # update the backend stats
        data = {}
        backend_name = self.configuration.safe_get('volume_backend_name')
        data['vendor_name'] = self.VENDOR
        data['driver_version'] = self.VERSION
        data['storage_protocol'] = self.PROTOCOL
        data['volume_backend_name'] = backend_name
        data['pools'] = self._get_pools_stats()

        self._stats = data
        LOG.debug('Update volume stats : %(stats)s.', {'stats': self._stats})

    def _build_target_portal(self, ip, port):
        """Build iSCSI portal for both IPV4 and IPV6."""
        addr = ipaddress.ip_address(ip)
        if addr.version == 4:
            ipaddr = ip
        else:
            ipaddr = '[%s]' % ip
        return '%(ip)s:%(port)s' % {'ip': ipaddr, 'port': port}

    @volume_utils.trace
    def initialize_connection(self, volume, connector, **kwargs):
        """Initialize connection steps:

        1. check if the host exist in targets.
        2.1 if there is target that has the host, add the volume to the target.
        2.2 if not, create an target add host to host add volume to host.
        3. return the target info.
        """
        host_ip = connector['ip']
        multipath = connector.get("multipath", False)
        # Check if there host exist in targets
        host_exist, target_name, node_of_target = self._get_target_from_conn(
            host_ip)
        if not host_exist:
            # host doesn't exist, need create target and bind the host,

            # generate the target name
            _TARGET_NAME_PATTERN = 'target.inspur.%(host)s-%(padding)s'
            _padding = str(random.randint(0, 99999999)).zfill(8)
            target_name = _TARGET_NAME_PATTERN % {'host': connector['host'],
                                                  'padding': _padding}

            # decide the nodes to be used
            if multipath:
                node_of_target = [node['name'] for node in self.nodes]
            else:
                # single node
                node_of_target = [self.nodes[0]['name']]

            # create the target
            nodes = ','.join(node_of_target)
            self._create_target(target_node=nodes,
                                target_name=target_name)
            self._add_host_to_target(host_ip=host_ip,
                                     target_name=target_name)

        self._add_lun_to_target(target_name=target_name, volume=volume)
        if self.configuration.use_chap_auth:
            self._add_chap_to_target(target_name,
                                     self.configuration.chap_username,
                                     self.configuration.chap_password)

        lun_id = self._get_lun_id(volume, target_name)
        connection_data = {
            'target_discovered': True,
            'volume_id': volume.id,
        }

        portals = []
        for node_name in node_of_target:
            for node in self.nodes:
                if node['name'] == node_name:
                    portal = self._build_target_portal(node.get('ip'), '3260')
                    portals.append(portal)

        if multipath:
            connection_data.update({
                'target_portals': portals,
                'target_luns': [int(lun_id)] * len(portals),
                'target_iqns': [target_name] * len(portals)
            })
        else:
            # single node
            connection_data.update({
                'target_portal': portals[0],
                'target_lun': int(lun_id),
                'target_iqn': target_name
            })

        if self.configuration.use_chap_auth:
            connection_data['auth_method'] = 'CHAP'
            connection_data['auth_username'] = self.configuration.chap_username
            connection_data['auth_password'] = self.configuration.chap_password

        return {'driver_volume_type': 'iscsi', 'data': connection_data}

    @volume_utils.trace
    def terminate_connection(self, volume, connector, **kwargs):
        """Delete lun from target.

        If target has no any lun, driver will delete the target.
        """
        volume_name = self._trans_name_down(volume.name)
        target_name = None
        lun_id = None

        host_ip = None
        if connector and 'ip' in connector:
            host_ip = connector['ip']

        target_list = self._get_target_list()
        for target in target_list:
            if not host_ip or host_ip in target['hostIp']:
                for lun in target['lun']:
                    if volume_name == lun['lvm']:
                        target_name = target['name']
                        lun_id = lun['lunID']
                        break
                if lun_id is not None:
                    break
        if lun_id is None:
            return

        self._delete_lun_from_target(target_name=target_name,
                                     lun_id=lun_id)
        luns = self._get_lun_list(target_name)
        if not luns:
            self._delete_target(target_name)

    def _get_pools_info(self, pools):
        """Get the pools info."""
        method = 'block/pool?type=2'
        requests_type = 'get'
        pools_data = self._rest.send_rest_api(method=method,
                                              request_type=requests_type)
        pools_info = {}
        for pool_data in pools_data:
            if pool_data['name'] in pools:
                pools_info[pool_data['name']] = pool_data

        return pools_info

    @volume_utils.trace
    def _get_pools_stats(self):
        """Generate the pool stat information."""
        pools_info = self._get_pools_info(self.pools)

        pools = []
        for pool_info in pools_info.values():
            total_capacity = pool_info.get('totalCapacity')
            total_capacity_gb = self._unit_convert(total_capacity)
            used_capacity = pool_info.get('usedCapacity')
            used_capacity_gb = self._unit_convert(used_capacity)
            free_capacity_gb = total_capacity_gb - used_capacity_gb

            pool = {
                'pool_name': pool_info.get('name'),
                'total_capacity_gb': total_capacity_gb,
                'free_capacity_gb': free_capacity_gb,
                'thin_provisioning_support': True,
                'thick_provisioning_support': False,
            }
            pools.append(pool)

        return pools

    @volume_utils.trace
    def _get_target_from_conn(self, host_ip):
        """Get target information base on the host ip."""
        host_exist = False
        target_name = None
        node = None

        target_list = self._get_target_list()
        for target in target_list:
            if host_ip in target['hostIp']:
                host_exist = True
                target_name = target['name']
                node = target['node']
                break

        return host_exist, target_name, node

    @volume_utils.trace
    def _get_target_list(self):
        """Get a list of all targets in the backend."""
        method = 'block/target/detail'
        request_type = 'get'
        data = self._rest.send_rest_api(method=method,
                                        request_type=request_type)
        return data

    @volume_utils.trace
    def _create_target(self, target_name, target_node):
        """Create a target on the specified node."""
        method = 'block/target'
        request_type = 'post'
        params = {'name': target_name, 'nodeName': target_node}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @volume_utils.trace
    def _delete_target(self, target_name):
        """Delete all target of all the node."""
        method = 'block/target?name=%s' % target_name
        request_type = 'delete'
        self._rest.send_rest_api(method=method,
                                 request_type=request_type)

    @volume_utils.trace
    def _add_chap_to_target(self, target_name, chap_username, chap_password):
        """Add CHAP to target."""
        method = 'block/chap/bond'
        request_type = 'post'
        params = {'target': target_name,
                  'user': chap_username,
                  'password': chap_password}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @volume_utils.trace
    def _add_host_to_target(self, host_ip, target_name):
        """Add the authority of host to target."""
        method = 'block/host'
        request_type = 'post'
        params = {'name': target_name, 'hostIp': host_ip}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @volume_utils.trace
    @utils.retry(retry_param=exception.VolumeDriverException,
                 interval=1,
                 retries=3)
    def _add_lun_to_target(self, target_name, volume):
        """Add volume to target."""
        pool = volume_utils.extract_host(volume.host, level='pool')
        volume_name = self._trans_name_down(volume.name)

        method = 'block/lun'
        request_type = 'post'
        params = {'name': target_name,
                  'pool': pool,
                  'lvm': volume_name}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @volume_utils.trace
    def _delete_lun_from_target(self, target_name, lun_id):
        """Delete lun from target_name."""
        method = 'block/lun?name=%s&id=%s&force=1' % (target_name, lun_id)
        request_type = 'delete'
        self._rest.send_rest_api(method=method, request_type=request_type)

    @volume_utils.trace
    def _get_lun_list(self, target_name):
        """Get all lun list of the target."""
        method = 'block/lun?name=%s' % target_name
        request_type = 'get'
        return self._rest.send_rest_api(method=method,
                                        request_type=request_type)

    @volume_utils.trace
    def _snapshot_lock_op(self, op, vol_name, snap_name, pool_name):
        """Lock or unlock a snapshot to protect the snapshot.

        op is 'lock' for lock and 'unlock' for unlock
        """
        method = 'snapshot/volume/%s' % op
        request_type = 'post'
        params = {'snapName': snap_name,
                  'volumeName': vol_name,
                  'poolName': pool_name}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @volume_utils.trace
    def _filling_volume(self, name, pool):
        """Filling a volume so that make it independently."""
        method = 'block/lvm/filling'
        request_type = 'post'
        params = {'pool': pool, 'name': name}
        self._rest.send_rest_api(method=method,
                                 params=params,
                                 request_type=request_type)

    @utils.retry(exception.VolumeDriverException, interval=5, retries=36)
    def _wait_volume_filled(self, name, pool):
        """Wait until the volume is filled."""
        volumes = self._get_volumes(pool)
        for vol in volumes:
            if name == vol['name']:
                if vol['lvmType'] == 1:
                    return
                else:
                    break
        msg = (_('Volume %s is not filled.') % name)
        raise exception.VolumeDriverException(msg)

    @volume_utils.trace
    def _check_volume(self, volume):
        """Check if the volume exists in the backend."""
        pool = volume_utils.extract_host(volume.host, 'pool')
        volume_name = self._trans_name_down(volume.name)
        attempts = 3
        while attempts > 0:
            volumes = self._get_volumes(pool)
            attempts -= 1
            for vol in volumes:
                if volume_name == vol.get('name'):
                    return True
            eventlet.sleep(1)
        return False

    @volume_utils.trace
    def _get_volumes(self, pool):
        """Get all the volumes in the pool."""
        method = 'block/lvm?pool=%s' % pool
        request_type = 'get'
        return self._rest.send_rest_api(method=method,
                                        request_type=request_type)

    @volume_utils.trace
    def _get_cluster_status(self):
        """Get all nodes of the backend."""
        method = 'cluster/node'
        request_type = 'get'
        return self._rest.send_rest_api(method=method,
                                        request_type=request_type)

    @volume_utils.trace
    def _get_lun_id(self, volume, target_name):
        """Get lun id of the voluem in a target."""
        pool = volume_utils.extract_host(volume.host, level='pool')
        volume_name = self._trans_name_down(volume.name)

        lun_id = None
        luns = self._get_lun_list(target_name)
        for lun in luns:
            mappinglvm = lun.get('mappingLvm')
            lun_name = mappinglvm.replace(r'%s/' % pool, '')
            if lun_name == volume_name:
                lun_id = lun.get('id')
        return lun_id

    def _trans_name_down(self, name):
        """Legitimize the name.

        Because AS13000 volume name is only allowed letters, numbers, and '_'.
        """
        return name.replace('-', '_')

    @volume_utils.trace
    def _unit_convert(self, capacity):
        """Convert all units to GB.

        The capacity is a string in form like 100GB, 20TB, 100B,
        this routine will convert to GB unit.
        """
        capacity = capacity.upper()
        try:
            unit = re.findall(r'[A-Z]+', capacity)[0]
        except BaseException:
            unit = ''
        capacity = float(capacity.replace(unit, ''))

        size_gb = 0.0

        if unit in ['B', '']:
            size_gb = capacity / units.Gi
        elif unit in ['K', 'KB']:
            size_gb = capacity / units.Mi
        elif unit in ['M', 'MB']:
            size_gb = capacity / units.Ki
        elif unit in ['G', 'GB']:
            size_gb = capacity
        elif unit in ['T', 'TB']:
            size_gb = capacity * units.Ki
        elif unit in ['P', 'PB']:
            size_gb = capacity * units.Mi
        elif unit in ['E', 'EB']:
            size_gb = capacity * units.Gi

        return float('%.0f' % size_gb)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Mock unit tests for NetApp Data ONTAP (C-mode) iSCSI storage systems."""

from unittest import mock


from cinder import context
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.netapp.dataontap import fakes as fake
import cinder.tests.unit.volume.drivers.netapp.fakes as na_fakes
from cinder.volume.drivers.netapp.dataontap import iscsi_cmode


class NetAppCmodeFibreChannelDriverTestCase(test.TestCase):

    def setUp(self):
        super(NetAppCmodeFibreChannelDriverTestCase, self).setUp()

        kwargs = {
            'configuration': self.get_config_base(),
            'host': 'openstack@netappblock',
        }
        self.library = iscsi_cmode.NetAppCmodeISCSIDriver(**kwargs)
        self.library.zapi_client = mock.Mock()
        self.zapi_client = self.library.zapi_client
        self.mock_request = mock.Mock()
        self.ctxt = context.RequestContext('fake', 'fake', auth_token=True)

    def get_config_base(self):
        return na_fakes.create_configuration()

    def test_revert_to_snapshot(self):
        mock_revert_to_snapshot = self.mock_object(self.library.library,
                                                   'revert_to_snapshot')

        self.library.revert_to_snapshot(self.ctxt, fake.SNAPSHOT_VOLUME,
                                        fake.SNAPSHOT)

        mock_revert_to_snapshot.assert_called_once_with(fake.SNAPSHOT_VOLUME,
                                                        fake.SNAPSHOT)



# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (c) 2014 TrilioData, Inc
# Copyright (c) 2015 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Handles all requests relating to the volume backups service."""

from datetime import datetime
import random

from eventlet import greenthread
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import strutils
from pytz import timezone

from cinder.backup import rpcapi as backup_rpcapi
from cinder.common import constants
from cinder import context
from cinder.db import base
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import fields
from cinder.policies import backup_actions as backup_action_policy
from cinder.policies import backups as policy
import cinder.policy
from cinder import quota
from cinder import quota_utils
from cinder.scheduler import rpcapi as scheduler_rpcapi
import cinder.volume

backup_opts = [
    cfg.BoolOpt('backup_use_same_host',
                default=False,
                help='Backup services use same backend.')
]

CONF = cfg.CONF
CONF.register_opts(backup_opts)
LOG = logging.getLogger(__name__)
QUOTAS = quota.QUOTAS
IMPORT_VOLUME_ID = '00000000-0000-0000-0000-000000000000'


class API(base.Base):
    """API for interacting with the volume backup manager."""

    def __init__(self):
        self.backup_rpcapi = backup_rpcapi.BackupAPI()
        self.scheduler_rpcapi = scheduler_rpcapi.SchedulerAPI()
        self.volume_api = cinder.volume.API()
        super().__init__()

    def get(self, context, backup_id):
        backup = objects.Backup.get_by_id(context, backup_id)
        context.authorize(policy.GET_POLICY, target_obj=backup)
        return backup

    def _check_support_to_force_delete(self, context, backup_host):
        result = self.backup_rpcapi.check_support_to_force_delete(context,
                                                                  backup_host)
        return result

    def delete(self, context, backup, force=False):
        """Make the RPC call to delete a volume backup.

        Call backup manager to execute backup delete or force delete operation.
        :param context: running context
        :param backup: the dict of backup that is got from DB.
        :param force: indicate force delete or not
        :raises InvalidBackup:
        :raises BackupDriverException:
        :raises ServiceNotFound:
        """
        context.authorize(policy.DELETE_POLICY, target_obj=backup)
        if not force and backup.status not in [fields.BackupStatus.AVAILABLE,
                                               fields.BackupStatus.ERROR]:
            msg = _('Backup status must be available or error')
            raise exception.InvalidBackup(reason=msg)
        if force and not self._check_support_to_force_delete(context,
                                                             backup.host):
            msg = _('force delete')
            raise exception.NotSupportedOperation(operation=msg)

        # Don't allow backup to be deleted if there are incremental
        # backups dependent on it.
        deltas = self.get_all(context, search_opts={'parent_id': backup.id})
        if deltas and len(deltas):
            msg = _('Incremental backups exist for this backup.')
            raise exception.InvalidBackup(reason=msg)

        backup.status = fields.BackupStatus.DELETING
        backup.host = self._get_available_backup_service_host(
            backup.host, backup.availability_zone)
        backup.save()
        self.backup_rpcapi.delete_backup(context, backup)

    def get_all(self, context, search_opts=None, marker=None, limit=None,
                offset=None, sort_keys=None, sort_dirs=None):
        context.authorize(policy.GET_ALL_POLICY)

        search_opts = search_opts or {}

        all_tenants = search_opts.pop('all_tenants', '0')
        if not strutils.is_valid_boolstr(all_tenants):
            msg = _("all_tenants must be a boolean, got '%s'.") % all_tenants
            raise exception.InvalidParameterValue(err=msg)

        if context.is_admin and strutils.bool_from_string(all_tenants):
            backups = objects.BackupList.get_all(context, search_opts,
                                                 marker, limit, offset,
                                                 sort_keys, sort_dirs)
        else:
            backups = objects.BackupList.get_all_by_project(
                context, context.project_id, search_opts,
                marker, limit, offset, sort_keys, sort_dirs
            )

        return backups

    def _az_matched(self, service, availability_zone):
        return ((not availability_zone) or
                service.availability_zone == availability_zone)

    def _is_backup_service_enabled(self, availability_zone, host):
        """Check if there is a backup service available."""
        topic = constants.BACKUP_TOPIC
        ctxt = context.get_admin_context()
        services = objects.ServiceList.get_all_by_topic(
            ctxt, topic, disabled=False)
        for srv in services:
            if (self._az_matched(srv, availability_zone) and
                    srv.host == host and srv.is_up):
                return True
        return False

    def _get_any_available_backup_service(self, availability_zone):
        """Get an available backup service host.

        Get an available backup service host in the specified
        availability zone.
        """
        services = [srv for srv in self._list_backup_services()]
        random.shuffle(services)
        # Get the next running service with matching availability zone.
        idx = 0
        while idx < len(services):
            srv = services[idx]
            if(self._az_matched(srv, availability_zone) and
               srv.is_up):
                return srv.host
            idx = idx + 1
        return None

    def get_available_backup_service_host(self, host, az):
        return self._get_available_backup_service_host(host, az)

    def _get_available_backup_service_host(self, host, az):
        """Return an appropriate backup service host."""
        backup_host = None
        if not host or not CONF.backup_use_same_host:
            backup_host = self._get_any_available_backup_service(az)
        elif self._is_backup_service_enabled(az, host):
            backup_host = host
        if not backup_host:
            raise exception.ServiceNotFound(service_id='cinder-backup')
        return backup_host

    def _list_backup_services(self):
        """List all enabled backup services.

        :returns: list -- hosts for services that are enabled for backup.
        """
        topic = constants.BACKUP_TOPIC
        ctxt = context.get_admin_context()
        services = objects.ServiceList.get_all_by_topic(
            ctxt, topic, disabled=False)
        return services

    def _list_backup_hosts(self):
        services = self._list_backup_services()
        return [srv.host for srv in services
                if not srv.disabled and srv.is_up]

    def create(self, context, name, description, volume_id,
               container, incremental=False, availability_zone=None,
               force=False, snapshot_id=None, metadata=None):
        """Make the RPC call to create a volume backup."""
        volume = self.volume_api.get(context, volume_id)
        context.authorize(policy.CREATE_POLICY, target_obj=volume)
        snapshot = None
        if snapshot_id:
            snapshot = self.volume_api.get_snapshot(context, snapshot_id)

            if volume_id != snapshot.volume_id:
                msg = (_('Volume %(vol1)s does not match with '
                         'snapshot.volume_id %(vol2)s.')
                       % {'vol1': volume_id,
                          'vol2': snapshot.volume_id})
                raise exception.InvalidVolume(reason=msg)
            if snapshot['status'] not in ["available"]:
                msg = (_('Snapshot to be backed up must be available, '
                         'but the current status is "%s".')
                       % snapshot['status'])
                raise exception.InvalidSnapshot(reason=msg)
        elif volume['status'] not in ["available", "in-use"]:
            msg = (_('Volume to be backed up must be available '
                     'or in-use, but the current status is "%s".')
                   % volume['status'])
            raise exception.InvalidVolume(reason=msg)
        elif volume['status'] in ["in-use"] and not force:
            msg = _('Backing up an in-use volume must use '
                    'the force flag.')
            raise exception.InvalidVolume(reason=msg)

        previous_status = volume['status']

        # Reserve a quota before setting volume status and backup status
        try:
            reserve_opts = {'backups': 1,
                            'backup_gigabytes': volume['size']}
            reservations = QUOTAS.reserve(context, **reserve_opts)
        except exception.OverQuota as e:
            quota_utils.process_reserve_over_quota(
                context, e,
                resource='backups',
                size=volume.size)
        # Find the latest backup and use it as the parent backup to do an
        # incremental backup.
        latest_backup = None
        if incremental:
            backups = objects.BackupList.get_all_by_volume(
                context, volume_id, volume['project_id'],
                filters={'project_id': context.project_id})
            if backups.objects:
                # NOTE(xyang): The 'data_timestamp' field records the time
                # when the data on the volume was first saved. If it is
                # a backup from volume, 'data_timestamp' will be the same
                # as 'created_at' for a backup. If it is a backup from a
                # snapshot, 'data_timestamp' will be the same as
                # 'created_at' for a snapshot.
                # If not backing up from snapshot, the backup with the latest
                # 'data_timestamp' will be the parent; If backing up from
                # snapshot, the backup with the latest 'data_timestamp' will
                # be chosen only if 'data_timestamp' is earlier than the
                # 'created_at' timestamp of the snapshot; Otherwise, the
                # backup will not be chosen as the parent.
                # For example, a volume has a backup taken at 8:00, then
                # a snapshot taken at 8:10, and then a backup at 8:20.
                # When taking an incremental backup of the snapshot, the
                # parent should be the backup at 8:00, not 8:20, and the
                # 'data_timestamp' of this new backup will be 8:10.
                latest_backup = max(
                    backups.objects,
                    key=lambda x: x['data_timestamp']
                    if (not snapshot or (snapshot and x['data_timestamp']
                                         < snapshot['created_at']))
                    else datetime(1, 1, 1, 1, 1, 1, tzinfo=timezone('UTC')))
            else:
                QUOTAS.rollback(context, reservations)
                msg = _('No backups available to do an incremental backup.')
                raise exception.InvalidBackup(reason=msg)

        parent_id = None
        parent = None

        if latest_backup:
            parent = latest_backup
            parent_id = latest_backup.id
            if latest_backup['status'] != fields.BackupStatus.AVAILABLE:
                QUOTAS.rollback(context, reservations)
                msg = _('The parent backup must be available for '
                        'incremental backup.')
                raise exception.InvalidBackup(reason=msg)

        data_timestamp = None
        if snapshot_id:
            snapshot = objects.Snapshot.get_by_id(context, snapshot_id)
            data_timestamp = snapshot.created_at
            self.db.snapshot_update(
                context, snapshot_id,
                {'status': fields.SnapshotStatus.BACKING_UP})
        else:
            self.db.volume_update(context, volume_id,
                                  {'status': 'backing-up',
                                   'previous_status': previous_status})

        backup = None
        try:
            kwargs = {
                'user_id': context.user_id,
                'project_id': context.project_id,
                'display_name': name,
                'display_description': description,
                'volume_id': volume_id,
                'status': fields.BackupStatus.CREATING,
                'container': container,
                'parent_id': parent_id,
                'size': volume['size'],
                'snapshot_id': snapshot_id,
                'data_timestamp': data_timestamp,
                'parent': parent,
                'metadata': metadata or {}
            }
            backup = objects.Backup(context=context, **kwargs)
            backup.create()
            if not snapshot_id:
                backup.data_timestamp = backup.created_at
                backup.save()
            QUOTAS.commit(context, reservations)
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    if backup and 'id' in backup:
                        backup.destroy()
                finally:
                    QUOTAS.rollback(context, reservations)

        self.scheduler_rpcapi.create_backup(context, backup)

        return backup

    def restore(self, context, backup_id, volume_id=None, name=None):
        """Make the RPC call to restore a volume backup."""
        backup = self.get(context, backup_id)
        context.authorize(policy.RESTORE_POLICY, target_obj=backup)
        if backup['status'] != fields.BackupStatus.AVAILABLE:
            msg = _('Backup status must be available')
            raise exception.InvalidBackup(reason=msg)

        size = backup['size']
        if size is None:
            msg = _('Backup to be restored has invalid size')
            raise exception.InvalidBackup(reason=msg)

        # Create a volume if none specified. If a volume is specified check
        # it is large enough for the backup
        if volume_id is None:
            if name is None:
                name = 'restore_backup_%s' % backup_id

            description = 'auto-created_from_restore_from_backup'

            LOG.info("Creating volume of %(size)s GB for restore of "
                     "backup %(backup_id)s.",
                     {'size': size, 'backup_id': backup_id})
            volume = self.volume_api.create(context, size, name, description)
            volume_id = volume['id']

            while True:
                volume = self.volume_api.get(context, volume_id)
                if volume['status'] != 'creating':
                    break
                greenthread.sleep(1)

            if volume['status'] == "error":
                msg = (_('Error while creating volume %(volume_id)s '
                         'for restoring backup %(backup_id)s.') %
                       {'volume_id': volume_id, 'backup_id': backup_id})
                raise exception.InvalidVolume(reason=msg)
        else:
            volume = self.volume_api.get(context, volume_id)

        if volume['status'] != "available":
            msg = _('Volume to be restored to must be available')
            raise exception.InvalidVolume(reason=msg)

        LOG.debug('Checking backup size %(bs)s against volume size %(vs)s',
                  {'bs': size, 'vs': volume['size']})
        if size > volume['size']:
            msg = (_('volume size %(volume_size)d is too small to restore '
                     'backup of size %(size)d.') %
                   {'volume_size': volume['size'], 'size': size})
            raise exception.InvalidVolume(reason=msg)

        LOG.info("Overwriting volume %(volume_id)s with restore of "
                 "backup %(backup_id)s",
                 {'volume_id': volume_id, 'backup_id': backup_id})

        # Setting the status here rather than setting at start and unrolling
        # for each error condition, it should be a very small window
        backup.host = self._get_available_backup_service_host(
            backup.host, backup.availability_zone)
        backup.status = fields.BackupStatus.RESTORING
        backup.restore_volume_id = volume.id
        backup.save()
        self.db.volume_update(context, volume_id, {'status':
                                                   'restoring-backup'})

        self.backup_rpcapi.restore_backup(context, backup.host, backup,
                                          volume_id)

        d = {'backup_id': backup_id,
             'volume_id': volume_id,
             'volume_name': volume['display_name'], }

        return d

    def reset_status(self, context, backup_id, status):
        """Make the RPC call to reset a volume backup's status.

        Call backup manager to execute backup status reset operation.
        :param context: running context
        :param backup_id: which backup's status to be reset
        :param status: backup's status to be reset
        :raises InvalidBackup:
        """
        # get backup info
        backup = self.get(context, backup_id)
        context.authorize(
            backup_action_policy.BASE_POLICY_NAME % "reset_status",
            target_obj=backup)
        backup.host = self._get_available_backup_service_host(
            backup.host, backup.availability_zone)
        backup.save()
        # send to manager to do reset operation
        self.backup_rpcapi.reset_status(ctxt=context, backup=backup,
                                        status=status)

    def export_record(self, context, backup_id):
        """Make the RPC call to export a volume backup.

        Call backup manager to execute backup export.

        :param context: running context
        :param backup_id: backup id to export
        :returns: dictionary -- a description of how to import the backup
        :returns: contains 'backup_url' and 'backup_service'
        :raises InvalidBackup:
        """
        backup = self.get(context, backup_id)
        context.authorize(policy.EXPORT_POLICY, target_obj=backup)
        if backup['status'] != fields.BackupStatus.AVAILABLE:
            msg = (_('Backup status must be available and not %s.') %
                   backup['status'])
            raise exception.InvalidBackup(reason=msg)

        LOG.debug("Calling RPCAPI with context: "
                  "%(ctx)s, host: %(host)s, backup: %(id)s.",
                  {'ctx': context,
                   'host': backup['host'],
                   'id': backup['id']})

        backup.host = self._get_available_backup_service_host(
            backup.host, backup.availability_zone)
        backup.save()
        export_data = self.backup_rpcapi.export_record(context, backup)

        return export_data

    def _get_import_backup(self, context, backup_url):
        """Prepare database backup record for import.

        This method decodes provided backup_url and expects to find the id of
        the backup in there.

        Then checks the DB for the presence of this backup record and if it
        finds it and is not deleted it will raise an exception because the
        record cannot be created or used.

        If the record is in deleted status then we must be trying to recover
        this record, so we'll reuse it.

        If the record doesn't already exist we create it with provided id.

        :param context: running context
        :param backup_url: backup description to be used by the backup driver
        :return: BackupImport object
        :raises InvalidBackup:
        :raises InvalidInput:
        """
        reservations = None
        backup = None
        # Deserialize string backup record into a dictionary
        backup_record = objects.Backup.decode_record(backup_url)

        # ID is a required field since it's what links incremental backups
        if 'id' not in backup_record:
            msg = _('Provided backup record is missing an id')
            raise exception.InvalidInput(reason=msg)

        # Since we use size to reserve&commit quota, size is another required
        # field.
        if 'size' not in backup_record:
            msg = _('Provided backup record is missing size attribute')
            raise exception.InvalidInput(reason=msg)

        try:
            reserve_opts = {'backups': 1,
                            'backup_gigabytes': backup_record['size']}
            reservations = QUOTAS.reserve(context, **reserve_opts)
        except exception.OverQuota as e:
            quota_utils.process_reserve_over_quota(
                context, e,
                resource='backups',
                size=backup_record['size'])

        kwargs = {
            'user_id': context.user_id,
            'project_id': context.project_id,
            'volume_id': IMPORT_VOLUME_ID,
            'status': fields.BackupStatus.CREATING,
            'deleted_at': None,
            'deleted': False,
            'metadata': {}
        }

        try:
            try:
                # Try to get the backup with that ID in all projects even among
                # deleted entries.
                backup = objects.BackupImport.get_by_id(
                    context.elevated(read_deleted='yes'),
                    backup_record['id'],
                    project_only=False)

                # If record exists and it's not deleted we cannot proceed
                # with the import
                if backup.status != fields.BackupStatus.DELETED:
                    msg = _('Backup already exists in database.')
                    raise exception.InvalidBackup(reason=msg)

                # Otherwise we'll "revive" delete backup record
                backup.update(kwargs)
                backup.save()
                QUOTAS.commit(context, reservations)
            except exception.BackupNotFound:
                # If record doesn't exist create it with the specific ID
                backup = objects.BackupImport(context=context,
                                              id=backup_record['id'], **kwargs)
                backup.create()
                QUOTAS.commit(context, reservations)
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    if backup and 'id' in backup:
                        backup.destroy()
                finally:
                    QUOTAS.rollback(context, reservations)
        return backup

    def import_record(self, context, backup_service, backup_url):
        """Make the RPC call to import a volume backup.

        :param context: running context
        :param backup_service: backup service name
        :param backup_url: backup description to be used by the backup driver
        :raises InvalidBackup:
        :raises ServiceNotFound:
        :raises InvalidInput:
        """
        context.authorize(policy.IMPORT_POLICY)

        # NOTE(ronenkat): since we don't have a backup-scheduler
        # we need to find a host that support the backup service
        # that was used to create the backup.
        # We  send it to the first backup service host, and the backup manager
        # on that host will forward it to other hosts on the hosts list if it
        # cannot support correct service itself.
        hosts = self._list_backup_hosts()
        if len(hosts) == 0:
            raise exception.ServiceNotFound(service_id=backup_service)

        # Get Backup object that will be used to import this backup record
        backup = self._get_import_backup(context, backup_url)

        first_host = hosts.pop()
        self.backup_rpcapi.import_record(context,
                                         first_host,
                                         backup,
                                         backup_service,
                                         backup_url,
                                         hosts)

        return backup

    def update(self, context, backup_id, fields):
        backup = self.get(context, backup_id)
        context.authorize(policy.UPDATE_POLICY, target_obj=backup)
        backup.update(fields)
        backup.save()
        return backup



# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Client side of the volume backup RPC API.
"""


from oslo_log import log as logging

from cinder.common import constants
from cinder import rpc


LOG = logging.getLogger(__name__)


class BackupAPI(rpc.RPCAPI):
    """Client side of the volume rpc API.

    API version history:

    .. code-block:: none

        1.0 - Initial version.
        1.1 - Changed methods to accept backup objects instead of IDs.
        1.2 - A version that got in by mistake (without breaking anything).
        1.3 - Dummy version bump to mark start of having cinder-backup service
              decoupled from cinder-volume.

        ... Mitaka supports messaging 1.3. Any changes to existing methods in
        1.x after this point should be done so that they can handle version cap
        set to 1.3.

        2.0 - Remove 1.x compatibility
        2.1 - Adds set_log_levels and get_log_levels
        2.2 - Adds publish_service_capabilities
    """

    RPC_API_VERSION = '2.2'
    RPC_DEFAULT_VERSION = '2.0'
    TOPIC = constants.BACKUP_TOPIC
    BINARY = 'cinder-backup'

    def create_backup(self, ctxt, backup):
        LOG.debug("create_backup in rpcapi backup_id %s", backup.id)
        cctxt = self._get_cctxt(server=backup.host)
        cctxt.cast(ctxt, 'create_backup', backup=backup)

    def restore_backup(self, ctxt, backup_host, backup, volume_id):
        LOG.debug("restore_backup in rpcapi backup_id %s", backup.id)
        cctxt = self._get_cctxt(server=backup_host)
        cctxt.cast(ctxt, 'restore_backup', backup=backup,
                   volume_id=volume_id)

    def delete_backup(self, ctxt, backup):
        LOG.debug("delete_backup rpcapi backup_id %s", backup.id)
        cctxt = self._get_cctxt(server=backup.host)
        cctxt.cast(ctxt, 'delete_backup', backup=backup)

    def export_record(self, ctxt, backup):
        LOG.debug("export_record in rpcapi backup_id %(id)s "
                  "on host %(host)s.",
                  {'id': backup.id,
                   'host': backup.host})
        cctxt = self._get_cctxt(server=backup.host)
        return cctxt.call(ctxt, 'export_record', backup=backup)

    def import_record(self, ctxt, host, backup, backup_service, backup_url,
                      backup_hosts):
        LOG.debug("import_record rpcapi backup id %(id)s "
                  "on host %(host)s for backup_url %(url)s.",
                  {'id': backup.id, 'host': host, 'url': backup_url})
        cctxt = self._get_cctxt(server=host)
        cctxt.cast(ctxt, 'import_record',
                   backup=backup,
                   backup_service=backup_service,
                   backup_url=backup_url,
                   backup_hosts=backup_hosts)

    def reset_status(self, ctxt, backup, status):
        LOG.debug("reset_status in rpcapi backup_id %(id)s "
                  "on host %(host)s.",
                  {'id': backup.id, 'host': backup.host})
        cctxt = self._get_cctxt(server=backup.host)
        return cctxt.cast(ctxt, 'reset_status', backup=backup, status=status)

    def check_support_to_force_delete(self, ctxt, host):
        LOG.debug("Check if backup driver supports force delete "
                  "on host %(host)s.", {'host': host})
        cctxt = self._get_cctxt(server=host)
        return cctxt.call(ctxt, 'check_support_to_force_delete')

    @rpc.assert_min_rpc_version('2.1')
    def set_log_levels(self, context, service, log_request):
        cctxt = self._get_cctxt(server=service.host, version='2.1')
        cctxt.cast(context, 'set_log_levels', log_request=log_request)

    @rpc.assert_min_rpc_version('2.1')
    def get_log_levels(self, context, service, log_request):
        cctxt = self._get_cctxt(server=service.host, version='2.1')
        return cctxt.call(context, 'get_log_levels', log_request=log_request)

    @rpc.assert_min_rpc_version('2.2')
    def publish_service_capabilities(self, ctxt):
        cctxt = self._get_cctxt(version='2.2', fanout=True)
        cctxt.cast(ctxt, 'publish_service_capabilities')



# Copyright (c) 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Volume-related Utilities and helpers."""

import abc
import ast
import functools
import inspect
import json
import logging as py_logging
import math
import operator
import os
from os import urandom
from random import shuffle
import re
import socket
import tempfile
import time
import types
import typing
from typing import Any, BinaryIO, Callable, Dict, IO  # noqa: H301
from typing import List, Optional, Tuple, Union  # noqa: H301
import uuid

from castellan.common.credentials import keystone_password
from castellan.common import exception as castellan_exception
from castellan import key_manager as castellan_key_manager
import eventlet
from eventlet import tpool
from keystoneauth1 import loading as ks_loading
from os_brick import encryptors
from os_brick.initiator import connector
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import netutils
from oslo_utils import strutils
from oslo_utils import timeutils
from oslo_utils import units

from cinder.brick.local_dev import lvm as brick_lvm
from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import objects
from cinder.objects import fields
from cinder import rpc
from cinder import utils
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume import group_types
from cinder.volume import throttling
from cinder.volume import volume_types

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

GB: int = units.Gi
# These attributes we will attempt to save for the volume if they exist
# in the source image metadata.
IMAGE_ATTRIBUTES = (
    'checksum',
    'container_format',
    'disk_format',
    'min_disk',
    'min_ram',
    'size',
)
VALID_TRACE_FLAGS = {'method', 'api'}
TRACE_API = False
TRACE_METHOD = False


def null_safe_str(s: Optional[str]) -> str:
    return str(s) if s else ''


def _usage_from_volume(context: context.RequestContext,
                       volume_ref: 'objects.Volume',
                       **kw) -> dict:
    now = timeutils.utcnow()
    launched_at = volume_ref['launched_at'] or now
    created_at = volume_ref['created_at'] or now
    volume_status = volume_ref['status']
    if volume_status == 'error_managing_deleting':
        volume_status = 'deleting'
    usage_info = dict(
        tenant_id=volume_ref['project_id'],
        host=volume_ref['host'],
        user_id=volume_ref['user_id'],
        availability_zone=volume_ref['availability_zone'],
        volume_id=volume_ref['id'],
        volume_type=volume_ref['volume_type_id'],
        display_name=volume_ref['display_name'],
        launched_at=launched_at.isoformat(),
        created_at=created_at.isoformat(),
        status=volume_status,
        snapshot_id=volume_ref['snapshot_id'],
        size=volume_ref['size'],
        replication_status=volume_ref['replication_status'],
        replication_extended_status=volume_ref['replication_extended_status'],
        replication_driver_data=volume_ref['replication_driver_data'],
        metadata=volume_ref.get('volume_metadata'),)

    usage_info.update(kw)
    try:
        attachments = db.volume_attachment_get_all_by_volume_id(
            context, volume_ref['id'])
        usage_info['volume_attachment'] = attachments

        glance_meta = db.volume_glance_metadata_get(context, volume_ref['id'])
        if glance_meta:
            usage_info['glance_metadata'] = glance_meta
    except exception.GlanceMetadataNotFound:
        pass
    except exception.VolumeNotFound:
        LOG.debug("Can not find volume %s at notify usage", volume_ref['id'])

    return usage_info


def _usage_from_backup(backup: 'objects.Backup', **kw) -> dict:
    num_dependent_backups = backup.num_dependent_backups
    usage_info = dict(tenant_id=backup.project_id,
                      user_id=backup.user_id,
                      availability_zone=backup.availability_zone,
                      backup_id=backup.id,
                      host=backup.host,
                      display_name=backup.display_name,
                      created_at=backup.created_at.isoformat(),
                      status=backup.status,
                      volume_id=backup.volume_id,
                      size=backup.size,
                      service_metadata=backup.service_metadata,
                      service=backup.service,
                      fail_reason=backup.fail_reason,
                      parent_id=backup.parent_id,
                      num_dependent_backups=num_dependent_backups,
                      snapshot_id=backup.snapshot_id,
                      )

    usage_info.update(kw)
    return usage_info


@utils.if_notifications_enabled
def notify_about_volume_usage(context: context.RequestContext,
                              volume: 'objects.Volume',
                              event_suffix: str,
                              extra_usage_info: dict = None,
                              host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_volume(context, volume, **extra_usage_info)

    rpc.get_notifier("volume", host).info(context, 'volume.%s' % event_suffix,
                                          usage_info)


@utils.if_notifications_enabled
def notify_about_backup_usage(context: context.RequestContext,
                              backup: 'objects.Backup',
                              event_suffix: str,
                              extra_usage_info: dict = None,
                              host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_backup(backup, **extra_usage_info)

    rpc.get_notifier("backup", host).info(context, 'backup.%s' % event_suffix,
                                          usage_info)


def _usage_from_snapshot(snapshot: 'objects.Snapshot',
                         context: context.RequestContext,
                         **extra_usage_info) -> dict:
    # (niedbalski) a snapshot might be related to a deleted
    # volume, if that's the case, the volume information is still
    # required for filling the usage_info, so we enforce to read
    # the volume data even if the volume has been deleted.
    context.read_deleted = "yes"
    volume = db.volume_get(context, snapshot.volume_id)
    usage_info = {
        'tenant_id': snapshot.project_id,
        'user_id': snapshot.user_id,
        'availability_zone': volume['availability_zone'],
        'volume_id': snapshot.volume_id,
        'volume_size': snapshot.volume_size,
        'snapshot_id': snapshot.id,
        'display_name': snapshot.display_name,
        'created_at': snapshot.created_at.isoformat(),
        'status': snapshot.status,
        'deleted': null_safe_str(snapshot.deleted),
        'metadata': null_safe_str(snapshot.metadata),
    }

    usage_info.update(extra_usage_info)
    return usage_info


@utils.if_notifications_enabled
def notify_about_snapshot_usage(context: context.RequestContext,
                                snapshot: 'objects.Snapshot',
                                event_suffix: str,
                                extra_usage_info: dict = None,
                                host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_snapshot(snapshot, context, **extra_usage_info)

    rpc.get_notifier('snapshot', host).info(context,
                                            'snapshot.%s' % event_suffix,
                                            usage_info)


def _usage_from_capacity(capacity: Dict[str, Any],
                         **extra_usage_info) -> Dict[str, Any]:

    capacity_info = {
        'name_to_id': capacity['name_to_id'],
        'total': capacity['total'],
        'free': capacity['free'],
        'allocated': capacity['allocated'],
        'provisioned': capacity['provisioned'],
        'virtual_free': capacity['virtual_free'],
        'reported_at': capacity['reported_at']
    }

    capacity_info.update(extra_usage_info)
    return capacity_info


@utils.if_notifications_enabled
def notify_about_capacity_usage(context: context.RequestContext,
                                capacity: dict,
                                suffix: str,
                                extra_usage_info: dict = None,
                                host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_capacity(capacity, **extra_usage_info)

    rpc.get_notifier('capacity', host).info(context,
                                            'capacity.%s' % suffix,
                                            usage_info)


@utils.if_notifications_enabled
def notify_about_replication_usage(context: context.RequestContext,
                                   volume: 'objects.Volume',
                                   suffix: str,
                                   extra_usage_info: dict = None,
                                   host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_volume(context, volume,
                                    **extra_usage_info)

    rpc.get_notifier('replication', host).info(context,
                                               'replication.%s' % suffix,
                                               usage_info)


@utils.if_notifications_enabled
def notify_about_replication_error(context: context.RequestContext,
                                   volume: 'objects.Volume',
                                   suffix: str,
                                   extra_error_info: dict = None,
                                   host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_error_info:
        extra_error_info = {}

    usage_info = _usage_from_volume(context, volume,
                                    **extra_error_info)

    rpc.get_notifier('replication', host).error(context,
                                                'replication.%s' % suffix,
                                                usage_info)


def _usage_from_consistencygroup(group_ref: 'objects.Group', **kw) -> dict:
    usage_info = dict(tenant_id=group_ref.project_id,
                      user_id=group_ref.user_id,
                      availability_zone=group_ref.availability_zone,
                      consistencygroup_id=group_ref.id,
                      name=group_ref.name,
                      created_at=group_ref.created_at.isoformat(),
                      status=group_ref.status)

    usage_info.update(kw)
    return usage_info


@utils.if_notifications_enabled
def notify_about_consistencygroup_usage(context: context.RequestContext,
                                        group: 'objects.Group',
                                        event_suffix: str,
                                        extra_usage_info: dict = None,
                                        host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_consistencygroup(group,
                                              **extra_usage_info)

    rpc.get_notifier("consistencygroup", host).info(
        context,
        'consistencygroup.%s' % event_suffix,
        usage_info)


def _usage_from_group(group_ref: 'objects.Group', **kw) -> dict:
    usage_info = dict(tenant_id=group_ref.project_id,
                      user_id=group_ref.user_id,
                      availability_zone=group_ref.availability_zone,
                      group_id=group_ref.id,
                      group_type=group_ref.group_type_id,
                      name=group_ref.name,
                      created_at=group_ref.created_at.isoformat(),
                      status=group_ref.status)

    usage_info.update(kw)
    return usage_info


@utils.if_notifications_enabled
def notify_about_group_usage(context: context.RequestContext,
                             group: 'objects.Group',
                             event_suffix: str,
                             extra_usage_info: dict = None,
                             host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_group(group,
                                   **extra_usage_info)

    rpc.get_notifier("group", host).info(
        context,
        'group.%s' % event_suffix,
        usage_info)


def _usage_from_cgsnapshot(cgsnapshot: 'objects.CGSnapshot', **kw) -> dict:
    usage_info = dict(
        tenant_id=cgsnapshot.project_id,
        user_id=cgsnapshot.user_id,
        cgsnapshot_id=cgsnapshot.id,
        name=cgsnapshot.name,
        consistencygroup_id=cgsnapshot.consistencygroup_id,
        created_at=cgsnapshot.created_at.isoformat(),
        status=cgsnapshot.status)

    usage_info.update(kw)
    return usage_info


def _usage_from_group_snapshot(group_snapshot: 'objects.GroupSnapshot',
                               **kw) -> dict:
    usage_info = dict(
        tenant_id=group_snapshot.project_id,
        user_id=group_snapshot.user_id,
        group_snapshot_id=group_snapshot.id,
        name=group_snapshot.name,
        group_id=group_snapshot.group_id,
        group_type=group_snapshot.group_type_id,
        created_at=group_snapshot.created_at.isoformat(),
        status=group_snapshot.status)

    usage_info.update(kw)
    return usage_info


@utils.if_notifications_enabled
def notify_about_cgsnapshot_usage(context: context.RequestContext,
                                  cgsnapshot: 'objects.CGSnapshot',
                                  event_suffix: str,
                                  extra_usage_info: dict = None,
                                  host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_cgsnapshot(cgsnapshot,
                                        **extra_usage_info)

    rpc.get_notifier("cgsnapshot", host).info(
        context,
        'cgsnapshot.%s' % event_suffix,
        usage_info)


@utils.if_notifications_enabled
def notify_about_group_snapshot_usage(context: context.RequestContext,
                                      group_snapshot: 'objects.GroupSnapshot',
                                      event_suffix: str,
                                      extra_usage_info=None,
                                      host: str = None) -> None:
    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_group_snapshot(group_snapshot,
                                            **extra_usage_info)

    rpc.get_notifier("group_snapshot", host).info(
        context,
        'group_snapshot.%s' % event_suffix,
        usage_info)


def _check_blocksize(blocksize: Union[str, int]) -> Union[str, int]:

    # Check if volume_dd_blocksize is valid
    try:
        # Rule out zero-sized/negative/float dd blocksize which
        # cannot be caught by strutils
        if (blocksize.startswith(('-', '0')) or  # type: ignore
                '.' in blocksize):  # type: ignore
            raise ValueError
        strutils.string_to_bytes('%sB' % blocksize)
    except ValueError:
        LOG.warning("Incorrect value error: %(blocksize)s, "
                    "it may indicate that \'volume_dd_blocksize\' "
                    "was configured incorrectly. Fall back to default.",
                    {'blocksize': blocksize})
        # Fall back to default blocksize
        CONF.clear_override('volume_dd_blocksize')
        blocksize = CONF.volume_dd_blocksize

    return blocksize


def check_for_odirect_support(src: str, dest: str,
                              flag: str = 'oflag=direct') -> bool:

    # Check whether O_DIRECT is supported
    try:
        # iflag=direct and if=/dev/zero combination does not work
        # error: dd: failed to open '/dev/zero': Invalid argument
        if (src == '/dev/zero' and flag == 'iflag=direct'):
            return False
        else:
            utils.execute('dd', 'count=0', 'if=%s' % src,
                          'of=%s' % dest,
                          flag, run_as_root=True)
            return True
    except processutils.ProcessExecutionError:
        return False


def _copy_volume_with_path(prefix, srcstr: str, deststr: str,
                           size_in_m: int, blocksize: Union[str, int],
                           sync: bool = False,
                           execute: Callable = utils.execute,
                           ionice=None,
                           sparse: bool = False) -> None:
    cmd = prefix[:]

    if ionice:
        cmd.extend(('ionice', ionice))

    blocksize = _check_blocksize(blocksize)
    size_in_bytes = size_in_m * units.Mi

    cmd.extend(('dd', 'if=%s' % srcstr, 'of=%s' % deststr,
                'count=%d' % size_in_bytes, 'bs=%s' % blocksize))

    # Use O_DIRECT to avoid thrashing the system buffer cache
    odirect = check_for_odirect_support(srcstr, deststr, 'iflag=direct')

    cmd.append('iflag=count_bytes,direct' if odirect else 'iflag=count_bytes')

    if check_for_odirect_support(srcstr, deststr, 'oflag=direct'):
        cmd.append('oflag=direct')
        odirect = True

    # If the volume is being unprovisioned then
    # request the data is persisted before returning,
    # so that it's not discarded from the cache.
    conv = []
    if sync and not odirect:
        conv.append('fdatasync')
    if sparse:
        conv.append('sparse')
    if conv:
        conv_options = 'conv=' + ",".join(conv)
        cmd.append(conv_options)

    # Perform the copy
    start_time = timeutils.utcnow()
    execute(*cmd, run_as_root=True)
    duration = timeutils.delta_seconds(start_time, timeutils.utcnow())

    # NOTE(jdg): use a default of 1, mostly for unit test, but in
    # some incredible event this is 0 (cirros image?) don't barf
    if duration < 1:
        duration = 1
    mbps = (size_in_m / duration)
    LOG.debug("Volume copy details: src %(src)s, dest %(dest)s, "
              "size %(sz).2f MB, duration %(duration).2f sec",
              {"src": srcstr,
               "dest": deststr,
               "sz": size_in_m,
               "duration": duration})
    LOG.info("Volume copy %(size_in_m).2f MB at %(mbps).2f MB/s",
             {'size_in_m': size_in_m, 'mbps': mbps})


def _open_volume_with_path(path: str, mode: str) -> IO[Any]:
    try:
        with utils.temporary_chown(path):
            handle = open(path, mode)
            return handle
    except Exception:
        LOG.error("Failed to open volume from %(path)s.", {'path': path})
        raise


def _transfer_data(src: IO, dest: IO,
                   length: int, chunk_size: int) -> None:
    """Transfer data between files (Python IO objects)."""

    chunks = int(math.ceil(length / chunk_size))
    remaining_length = length

    LOG.debug("%(chunks)s chunks of %(bytes)s bytes to be transferred.",
              {'chunks': chunks, 'bytes': chunk_size})

    for chunk in range(0, chunks):
        before = time.time()
        data = tpool.execute(src.read, min(chunk_size, remaining_length))

        # If we have reached end of source, discard any extraneous bytes from
        # destination volume if trim is enabled and stop writing.
        if data == b'':
            break

        tpool.execute(dest.write, data)
        remaining_length -= len(data)
        delta = (time.time() - before)
        rate = (chunk_size / delta) / units.Ki
        LOG.debug("Transferred chunk %(chunk)s of %(chunks)s (%(rate)dK/s).",
                  {'chunk': chunk + 1, 'chunks': chunks, 'rate': rate})

        # yield to any other pending operations
        eventlet.sleep(0)

    tpool.execute(dest.flush)


def _copy_volume_with_file(src: Union[str, IO],
                           dest: Union[str, IO],
                           size_in_m: int) -> None:
    src_handle = src
    if isinstance(src, str):
        src_handle = _open_volume_with_path(src, 'rb')

    src_handle = typing.cast(IO, src_handle)

    dest_handle = dest
    if isinstance(dest, str):
        dest_handle = _open_volume_with_path(dest, 'wb')

    dest_handle = typing.cast(IO, dest_handle)

    if not src_handle:
        raise exception.DeviceUnavailable(
            _("Failed to copy volume, source device unavailable."))

    if not dest_handle:
        raise exception.DeviceUnavailable(
            _("Failed to copy volume, destination device unavailable."))

    start_time = timeutils.utcnow()

    _transfer_data(src_handle, dest_handle, size_in_m * units.Mi, units.Mi * 4)

    duration = max(1, timeutils.delta_seconds(start_time, timeutils.utcnow()))

    if isinstance(src, str):
        src_handle.close()
    if isinstance(dest, str):
        dest_handle.close()

    mbps = (size_in_m / duration)
    LOG.info("Volume copy completed (%(size_in_m).2f MB at "
             "%(mbps).2f MB/s).",
             {'size_in_m': size_in_m, 'mbps': mbps})


def copy_volume(src: Union[str, BinaryIO],
                dest: Union[str, BinaryIO],
                size_in_m: int,
                blocksize: Union[str, int], sync=False,
                execute=utils.execute, ionice=None, throttle=None,
                sparse=False) -> None:
    """Copy data from the source volume to the destination volume.

    The parameters 'src' and 'dest' are both typically of type str, which
    represents the path to each volume on the filesystem.  Connectors can
    optionally return a volume handle of type RawIOBase for volumes that are
    not available on the local filesystem for open/close operations.

    If either 'src' or 'dest' are not of type str, then they are assumed to be
    of type RawIOBase or any derivative that supports file operations such as
    read and write.  In this case, the handles are treated as file handles
    instead of file paths and, at present moment, throttling is unavailable.
    """

    if (isinstance(src, str) and
            isinstance(dest, str)):
        if not throttle:
            throttle = throttling.Throttle.get_default()
        with throttle.subcommand(src, dest) as throttle_cmd:
            _copy_volume_with_path(throttle_cmd['prefix'], src, dest,
                                   size_in_m, blocksize, sync=sync,
                                   execute=execute, ionice=ionice,
                                   sparse=sparse)
    else:
        _copy_volume_with_file(src, dest, size_in_m)


def clear_volume(volume_size: int,
                 volume_path: str,
                 volume_clear: str = None,
                 volume_clear_size: int = None,
                 volume_clear_ionice: str = None,
                 throttle=None) -> None:
    """Unprovision old volumes to prevent data leaking between users."""
    if volume_clear is None:
        volume_clear = CONF.volume_clear

    if volume_clear_size is None:
        volume_clear_size = CONF.volume_clear_size

    if volume_clear_size == 0:
        volume_clear_size = volume_size

    if volume_clear_ionice is None:
        volume_clear_ionice = CONF.volume_clear_ionice

    LOG.info("Performing secure delete on volume: %s", volume_path)

    # We pass sparse=False explicitly here so that zero blocks are not
    # skipped in order to clear the volume.
    if volume_clear == 'zero':
        return copy_volume('/dev/zero', volume_path, volume_clear_size,
                           CONF.volume_dd_blocksize,
                           sync=True, execute=utils.execute,
                           ionice=volume_clear_ionice,
                           throttle=throttle, sparse=False)
    else:
        raise exception.InvalidConfigurationValue(
            option='volume_clear',
            value=volume_clear)


def supports_thin_provisioning() -> bool:
    return brick_lvm.LVM.supports_thin_provisioning(
        utils.get_root_helper())


def get_all_physical_volumes(vg_name=None) -> list:
    return brick_lvm.LVM.get_all_physical_volumes(
        utils.get_root_helper(),
        vg_name)


def get_all_volume_groups(vg_name=None) -> list:
    return brick_lvm.LVM.get_all_volume_groups(
        utils.get_root_helper(),
        vg_name)


def extract_availability_zones_from_volume_type(
        volume_type: Union['objects.VolumeType', dict]) \
        -> Optional[List[str]]:
    if not volume_type:
        return None
    extra_specs = volume_type.get('extra_specs', {})
    if 'RESKEY:availability_zones' not in extra_specs:
        return None
    azs = extra_specs.get('RESKEY:availability_zones', '').split(',')
    return [az.strip() for az in azs if az != '']


# Default symbols to use for passwords. Avoids visually confusing characters.
# ~6 bits per symbol
DEFAULT_PASSWORD_SYMBOLS = ('23456789',  # Removed: 0,1
                            'ABCDEFGHJKLMNPQRSTUVWXYZ',   # Removed: I, O
                            'abcdefghijkmnopqrstuvwxyz')  # Removed: l


def generate_password(
        length: int = 16,
        symbolgroups: Tuple[str, ...] = DEFAULT_PASSWORD_SYMBOLS) -> str:
    """Generate a random password from the supplied symbol groups.

    At least one symbol from each group will be included. Unpredictable
    results if length is less than the number of symbol groups.

    Believed to be reasonably secure (with a reasonable password length!)

    """
    # NOTE(jerdfelt): Some password policies require at least one character
    # from each group of symbols, so start off with one random character
    # from each symbol group

    bytes = 1  # Number of random bytes to generate for each choice

    password = [s[ord(urandom(bytes)) % len(s)]
                for s in symbolgroups]
    # If length < len(symbolgroups), the leading characters will only
    # be from the first length groups. Try our best to not be predictable
    # by shuffling and then truncating.
    shuffle(password)
    password = password[:length]
    length -= len(password)

    # then fill with random characters from all symbol groups
    symbols = ''.join(symbolgroups)
    password.extend(
        [symbols[ord(urandom(bytes)) % len(symbols)]
            for _i in range(length)])

    # finally shuffle to ensure first x characters aren't from a
    # predictable group
    shuffle(password)

    return ''.join(password)


def generate_username(
        length: int = 20,
        symbolgroups: Tuple[str, ...] = DEFAULT_PASSWORD_SYMBOLS) -> str:
    # Use the same implementation as the password generation.
    return generate_password(length, symbolgroups)


DEFAULT_POOL_NAME = '_pool0'


def extract_host(host: Optional[str],
                 level: str = 'backend',
                 default_pool_name: bool = False) -> Optional[str]:
    """Extract Host, Backend or Pool information from host string.

    :param host: String for host, which could include host@backend#pool info
    :param level: Indicate which level of information should be extracted
                  from host string. Level can be 'host', 'backend' or 'pool',
                  default value is 'backend'
    :param default_pool_name: this flag specify what to do if level == 'pool'
                              and there is no 'pool' info encoded in host
                              string.  default_pool_name=True will return
                              DEFAULT_POOL_NAME, otherwise we return None.
                              Default value of this parameter is False.
    :return: expected information, string or None
    :raises: exception.InvalidVolume

    For example:
        host = 'HostA@BackendB#PoolC'
        ret = extract_host(host, 'host')
        # ret is 'HostA'
        ret = extract_host(host, 'backend')
        # ret is 'HostA@BackendB'
        ret = extract_host(host, 'pool')
        # ret is 'PoolC'

        host = 'HostX@BackendY'
        ret = extract_host(host, 'pool')
        # ret is None
        ret = extract_host(host, 'pool', True)
        # ret is '_pool0'
    """

    if host is None:
        msg = _("volume is not assigned to a host")
        raise exception.InvalidVolume(reason=msg)

    if level == 'host':
        # make sure pool is not included
        hst = host.split('#')[0]
        return hst.split('@')[0]
    elif level == 'backend':
        return host.split('#')[0]
    elif level == 'pool':
        lst = host.split('#')
        if len(lst) == 2:
            return lst[1]
        elif default_pool_name is True:
            return DEFAULT_POOL_NAME
        else:
            return None

    return None  # not hit


def append_host(host: Optional[str],
                pool: Optional[str]) -> Optional[str]:
    """Encode pool into host info."""
    if not host or not pool:
        return host

    new_host = "#".join([host, pool])
    return new_host


def matching_backend_name(src_volume_type, volume_type) -> bool:
    if src_volume_type.get('volume_backend_name') and \
            volume_type.get('volume_backend_name'):
        return src_volume_type.get('volume_backend_name') == \
            volume_type.get('volume_backend_name')
    else:
        return False


def hosts_are_equivalent(host_1: str, host_2: str) -> bool:
    # In case host_1 or host_2 are None
    if not (host_1 and host_2):
        return host_1 == host_2
    return extract_host(host_1) == extract_host(host_2)


def read_proc_mounts() -> List[str]:
    """Read the /proc/mounts file.

    It's a dummy function but it eases the writing of unit tests as mocking
    __builtin__open() for a specific file only is not trivial.
    """
    with open('/proc/mounts') as mounts:
        return mounts.readlines()


def extract_id_from_volume_name(vol_name: str) -> Optional[str]:
    regex: typing.Pattern = re.compile(
        CONF.volume_name_template.replace('%s', r'(?P<uuid>.+)'))
    match = regex.match(vol_name)
    return match.group('uuid') if match else None


def check_already_managed_volume(vol_id: Optional[str]):
    """Check cinder db for already managed volume.

    :param vol_id: volume id parameter
    :returns: bool -- return True, if db entry with specified
                      volume id exists, otherwise return False
    :raises: ValueError if vol_id is not a valid uuid string
    """
    try:
        return (vol_id and isinstance(vol_id, str) and
                uuid.UUID(vol_id, version=4) and
                objects.Volume.exists(context.get_admin_context(), vol_id))
    except ValueError:
        return False


def extract_id_from_snapshot_name(snap_name: str) -> Optional[str]:
    """Return a snapshot's ID from its name on the backend."""
    regex = re.compile(
        CONF.snapshot_name_template.replace('%s', r'(?P<uuid>.+)'))
    match = regex.match(snap_name)
    return match.group('uuid') if match else None


def paginate_entries_list(entries: List[Dict],
                          marker: Optional[Union[dict, str]],
                          limit: int,
                          offset: Optional[int],
                          sort_keys: List[str],
                          sort_dirs: List[str]) -> list:
    """Paginate a list of entries.

    :param entries: list of dictionaries
    :marker: The last element previously returned
    :limit: The maximum number of items to return
    :offset: The number of items to skip from the marker or from the first
             element.
    :sort_keys: A list of keys in the dictionaries to sort by
    :sort_dirs: A list of sort directions, where each is either 'asc' or 'dec'
    """
    comparers = [(operator.itemgetter(key.strip()), multiplier)
                 for (key, multiplier) in zip(sort_keys, sort_dirs)]

    def comparer(left, right) -> int:
        fn: Callable
        for fn, d in comparers:
            left_val = fn(left)
            right_val = fn(right)
            if isinstance(left_val, dict):
                left_val = sorted(left_val.values())[0]
            if isinstance(right_val, dict):
                right_val = sorted(right_val.values())[0]
            if left_val == right_val:
                continue
            if d == 'asc':
                return -1 if left_val < right_val else 1
            else:
                return -1 if left_val > right_val else 1
        else:
            return 0
    sorted_entries = sorted(entries, key=functools.cmp_to_key(comparer))

    start_index = 0
    if offset is None:
        offset = 0
    if marker:
        if not isinstance(marker, dict):
            try:
                marker = json.loads(marker)
            except ValueError:
                msg = _('marker %s can not be analysed, please use json like '
                        'format') % marker
                raise exception.InvalidInput(reason=msg)
        start_index = -1
        for i, entry in enumerate(sorted_entries):
            if entry['reference'] == marker:
                start_index = i + 1
                break
        if start_index < 0:
            msg = _('marker not found: %s') % marker
            raise exception.InvalidInput(reason=msg)
    range_end = start_index + limit
    return sorted_entries[start_index + offset:range_end + offset]


def convert_config_string_to_dict(config_string: str) -> dict:
    """Convert config file replication string to a dict.

    The only supported form is as follows:
    "{'key-1'='val-1' 'key-2'='val-2'...}"

    :param config_string: Properly formatted string to convert to dict.
    :response: dict of string values
    """

    resultant_dict = {}

    try:
        st = config_string.replace("=", ":")
        st = st.replace(" ", ", ")
        resultant_dict = ast.literal_eval(st)
    except Exception:
        LOG.warning("Error encountered translating config_string: "
                    "%(config_string)s to dict",
                    {'config_string': config_string})

    return resultant_dict


def create_encryption_key(context: context.RequestContext,
                          key_manager,
                          volume_type_id: str) -> Optional[str]:
    encryption_key_id = None
    if volume_types.is_encrypted(context, volume_type_id):
        volume_type_encryption: db.sqlalchemy.models.Encryption = (
            volume_types.get_volume_type_encryption(context,
                                                    volume_type_id))
        if volume_type_encryption is None:
            raise exception.Invalid(message="Volume type error")
        cipher = volume_type_encryption.cipher
        length = volume_type_encryption.key_size
        algorithm = cipher.split('-')[0] if cipher else None
        if algorithm is None:
            raise exception.InvalidVolumeType(
                message="Invalid encryption spec")
        try:
            encryption_key_id = key_manager.create_key(
                context,
                algorithm=algorithm,
                length=length)
        except castellan_exception.KeyManagerError:
            # The messaging back to the client here is
            # purposefully terse, so we don't leak any sensitive
            # details.
            LOG.exception("Key manager error")
            raise exception.Invalid(message="Key manager error")

    typing.cast(str, encryption_key_id)

    return encryption_key_id


def delete_encryption_key(context: context.RequestContext,
                          key_manager,
                          encryption_key_id: str) -> None:
    try:
        key_manager.delete(context, encryption_key_id)
    except castellan_exception.ManagedObjectNotFoundError:
        pass
    except castellan_exception.KeyManagerError:
        LOG.info("First attempt to delete key id %s failed, retrying with "
                 "cinder's service context.", encryption_key_id)
        conf = CONF
        ks_loading.register_auth_conf_options(conf, 'keystone_authtoken')
        service_context = keystone_password.KeystonePassword(
            password=conf.keystone_authtoken.password,
            auth_url=conf.keystone_authtoken.auth_url,
            username=conf.keystone_authtoken.username,
            user_domain_name=conf.keystone_authtoken.user_domain_name,
            project_name=conf.keystone_authtoken.project_name,
            project_domain_name=conf.keystone_authtoken.project_domain_name)
        try:
            castellan_key_manager.API(conf).delete(service_context,
                                                   encryption_key_id)
        except castellan_exception.ManagedObjectNotFoundError:
            pass


def clone_encryption_key(context: context.RequestContext,
                         key_manager,
                         encryption_key_id: str) -> str:
    clone_key_id = None
    if encryption_key_id is not None:
        clone_key_id = key_manager.store(
            context,
            key_manager.get(context, encryption_key_id))
    return clone_key_id


def is_boolean_str(str: Optional[str]) -> bool:
    spec = (str or '').split()
    return (len(spec) == 2 and
            spec[0] == '<is>' and strutils.bool_from_string(spec[1]))


def is_replicated_spec(extra_specs: dict) -> bool:
    return (bool(extra_specs) and
            is_boolean_str(extra_specs.get('replication_enabled')))


def is_multiattach_spec(extra_specs: dict) -> bool:
    return (bool(extra_specs) and
            is_boolean_str(extra_specs.get('multiattach')))


def group_get_by_id(group_id):
    ctxt = context.get_admin_context()
    group = db.group_get(ctxt, group_id)
    return group


def is_group_a_cg_snapshot_type(group_or_snap) -> bool:
    LOG.debug("Checking if %s is a consistent snapshot group",
              group_or_snap)
    if group_or_snap["group_type_id"] is not None:
        spec = group_types.get_group_type_specs(
            group_or_snap["group_type_id"],
            key="consistent_group_snapshot_enabled"
        )
        return spec == "<is> True"
    return False


def is_group_a_type(group: 'objects.Group', key: str) -> bool:
    if group.group_type_id is not None:
        spec = group_types.get_group_type_specs(
            group.group_type_id, key=key
        )
        return spec == "<is> True"
    return False


def get_max_over_subscription_ratio(
        str_value: Union[str, float],
        supports_auto: bool = False) -> Union[str, float]:
    """Get the max_over_subscription_ratio from a string

    As some drivers need to do some calculations with the value and we are now
    receiving a string value in the conf, this converts the value to float
    when appropriate.

    :param str_value: Configuration object
    :param supports_auto: Tell if the calling driver supports auto MOSR.
    :response: value of mosr
    """

    if not supports_auto and str_value == "auto":
        msg = _("This driver does not support automatic "
                "max_over_subscription_ratio calculation. Please use a "
                "valid float value.")
        LOG.error(msg)
        raise exception.VolumeDriverException(message=msg)

    if str_value == 'auto':
        str_value = typing.cast(str, str_value)
        return str_value

    mosr = float(str_value)
    if mosr < 1:
        msg = _("The value of max_over_subscription_ratio must be "
                "greater than 1.")
        LOG.error(msg)
        raise exception.InvalidParameterValue(message=msg)
    return mosr


def check_image_metadata(image_meta: Dict[str, Union[str, int]],
                         vol_size: int) -> None:
    """Validates the image metadata."""
    # Check whether image is active
    if image_meta['status'] != 'active':
        msg = _('Image %(image_id)s is not active.'
                ) % {'image_id': image_meta['id']}
        raise exception.InvalidInput(reason=msg)

    # Check image size is not larger than volume size.
    image_size = utils.as_int(image_meta['size'], quiet=False)
    image_size_in_gb = (image_size + GB - 1) // GB
    if image_size_in_gb > vol_size:
        msg = _('Size of specified image %(image_size)sGB'
                ' is larger than volume size %(volume_size)sGB.')
        msg = msg % {'image_size': image_size_in_gb, 'volume_size': vol_size}
        raise exception.InvalidInput(reason=msg)

    # Check image min_disk requirement is met for the particular volume
    min_disk = image_meta.get('min_disk', 0)
    min_disk = typing.cast(int, min_disk)
    if vol_size < min_disk:
        msg = _('Volume size %(volume_size)sGB cannot be smaller'
                ' than the image minDisk size %(min_disk)sGB.')
        msg = msg % {'volume_size': vol_size, 'min_disk': min_disk}
        raise exception.InvalidInput(reason=msg)


def enable_bootable_flag(volume: 'objects.Volume') -> None:
    try:
        LOG.debug('Marking volume %s as bootable.', volume.id)
        volume.bootable = True
        volume.save()
    except exception.CinderException as ex:
        LOG.exception("Failed updating volume %(volume_id)s bootable "
                      "flag to true", {'volume_id': volume.id})
        raise exception.MetadataUpdateFailure(reason=ex)


def get_volume_image_metadata(image_id: str,
                              image_meta: Dict[str, Any]) -> dict:

    # Save some base attributes into the volume metadata
    base_metadata = {
        'image_id': image_id,
    }
    name = image_meta.get('name', None)
    if name:
        base_metadata['image_name'] = name

    # Save some more attributes into the volume metadata from the image
    # metadata
    for key in IMAGE_ATTRIBUTES:
        if key not in image_meta:
            continue
        value = image_meta.get(key, None)
        if value is not None:
            base_metadata[key] = value

    # Save all the image metadata properties into the volume metadata
    property_metadata = {}
    image_properties = image_meta.get('properties', {})
    image_properties = typing.cast(dict, image_properties)
    for (key, value) in image_properties.items():
        if value is not None:
            property_metadata[key] = value

    volume_metadata = dict(property_metadata)
    volume_metadata.update(base_metadata)
    return volume_metadata


def copy_image_to_volume(driver,
                         context: context.RequestContext,
                         volume: 'objects.Volume',
                         image_meta: dict,
                         image_location: str,
                         image_service) -> None:
    """Downloads Glance image to the specified volume."""
    image_id = image_meta['id']
    LOG.debug("Attempting download of %(image_id)s (%(image_location)s)"
              " to volume %(volume_id)s.",
              {'image_id': image_id, 'volume_id': volume.id,
               'image_location': image_location})
    try:
        image_encryption_key = image_meta.get('cinder_encryption_key_id')

        if volume.encryption_key_id and image_encryption_key:
            # If the image provided an encryption key, we have
            # already cloned it to the volume's key in
            # _get_encryption_key_id, so we can do a direct copy.
            driver.copy_image_to_volume(
                context, volume, image_service, image_id)
        elif volume.encryption_key_id:
            # Creating an encrypted volume from a normal, unencrypted,
            # image.
            driver.copy_image_to_encrypted_volume(
                context, volume, image_service, image_id)
        else:
            driver.copy_image_to_volume(
                context, volume, image_service, image_id)
    except processutils.ProcessExecutionError as ex:
        LOG.exception("Failed to copy image %(image_id)s to volume: "
                      "%(volume_id)s",
                      {'volume_id': volume.id, 'image_id': image_id})
        raise exception.ImageCopyFailure(reason=ex.stderr)
    except (exception.ImageUnacceptable, exception.ImageTooBig):
        with excutils.save_and_reraise_exception():
            LOG.exception("Failed to copy image %(image_id)s to volume: "
                          "%(volume_id)s",
                          {'volume_id': volume.id, 'image_id': image_id})
    except Exception as ex:
        LOG.exception("Failed to copy image %(image_id)s to "
                      "volume: %(volume_id)s",
                      {'volume_id': volume.id, 'image_id': image_id})
        if not isinstance(ex, exception.ImageCopyFailure):
            raise exception.ImageCopyFailure(reason=ex)
        else:
            raise

    LOG.debug("Downloaded image %(image_id)s (%(image_location)s)"
              " to volume %(volume_id)s successfully.",
              {'image_id': image_id, 'volume_id': volume.id,
               'image_location': image_location})


def image_conversion_dir() -> str:
    tmpdir = (CONF.image_conversion_dir or
              tempfile.gettempdir())

    # ensure temporary directory exists
    if not os.path.exists(tmpdir):
        os.makedirs(tmpdir)

    return tmpdir


def check_encryption_provider(
    volume: 'objects.Volume',
    context: context.RequestContext,
) -> dict:
    """Check that this is a LUKS encryption provider.

    :returns: encryption dict
    """

    encryption = db.volume_encryption_metadata_get(context, volume.id)

    if 'provider' not in encryption:
        message = _("Invalid encryption spec.")
        raise exception.VolumeDriverException(message=message)

    provider = encryption['provider']
    if provider in encryptors.LEGACY_PROVIDER_CLASS_TO_FORMAT_MAP:
        provider = encryptors.LEGACY_PROVIDER_CLASS_TO_FORMAT_MAP[provider]
        encryption['provider'] = provider
    if provider != encryptors.LUKS:
        message = _("Provider %s not supported.") % provider
        raise exception.VolumeDriverException(message=message)

    if 'cipher' not in encryption or 'key_size' not in encryption:
        msg = _('encryption spec must contain "cipher" and '
                '"key_size"')
        raise exception.VolumeDriverException(message=msg)

    return encryption


def sanitize_host(host: str) -> str:
    """Ensure IPv6 addresses are enclosed in [] for iSCSI portals."""
    if netutils.is_valid_ipv6(host):
        return '[%s]' % host
    return host


def sanitize_hostname(hostname) -> str:
    """Return a hostname which conforms to RFC-952 and RFC-1123 specs."""
    hostname = hostname.encode('latin-1', 'ignore')
    hostname = hostname.decode('latin-1')

    hostname = re.sub(r'[ _]', '-', hostname)
    hostname = re.sub(r'[^\w.-]+', '', hostname)
    hostname = hostname.lower()
    hostname = hostname.strip('.-')

    return hostname


def resolve_hostname(hostname: str) -> str:
    """Resolves host name to IP address.

    Resolves a host name (my.data.point.com) to an IP address (10.12.143.11).
    This routine also works if the data passed in hostname is already an IP.
    In this case, the same IP address will be returned.

    :param hostname:  Host name to resolve.
    :returns:         IP Address for Host name.
    """
    ip = socket.getaddrinfo(hostname, None)[0][4][0]
    LOG.debug('Asked to resolve hostname %(host)s and got IP %(ip)s.',
              {'host': hostname, 'ip': ip})
    return ip


def update_backup_error(backup,
                        err: str,
                        status=fields.BackupStatus.ERROR) -> None:
    backup.status = status
    backup.fail_reason = err
    backup.save()


# TODO (whoami-rajat): Remove this method when oslo.vmware calls volume_utils
#  wrapper of upload_volume instead of image_utils.upload_volume
def get_base_image_ref(volume: 'objects.Volume'):
    # This method fetches the image_id from volume glance metadata and pass
    # it to the driver calling it during upload volume to image operation
    base_image_ref = None
    if volume.glance_metadata:
        base_image_ref = volume.glance_metadata.get('image_id')
    return base_image_ref


def upload_volume(context: context.RequestContext,
                  image_service, image_meta, volume_path,
                  volume: 'objects.Volume',
                  volume_format: str = 'raw',
                  run_as_root: bool = True,
                  compress: bool = True) -> None:
    # retrieve store information from extra-specs
    store_id = volume.volume_type.extra_specs.get('image_service:store_id')

    # This fetches the image_id from volume glance metadata and pass
    # it to the driver calling it during upload volume to image operation
    base_image_ref = None
    if volume.glance_metadata:
        base_image_ref = volume.glance_metadata.get('image_id')
    image_utils.upload_volume(context, image_service, image_meta, volume_path,
                              volume_format=volume_format,
                              run_as_root=run_as_root,
                              compress=compress, store_id=store_id,
                              base_image_ref=base_image_ref)


def get_backend_configuration(backend_name, backend_opts=None):
    """Get a configuration object for a specific backend."""

    config_stanzas = CONF.list_all_sections()
    if backend_name not in config_stanzas:
        msg = _("Could not find backend stanza %(backend_name)s in "
                "configuration. Available stanzas are %(stanzas)s")
        params = {
            "stanzas": config_stanzas,
            "backend_name": backend_name,
        }
        raise exception.ConfigNotFound(message=msg % params)

    config = configuration.Configuration(driver.volume_opts,
                                         config_group=backend_name)

    if backend_opts:
        config.append_config_values(backend_opts)

    return config


def brick_get_connector_properties(multipath: bool = False,
                                   enforce_multipath: bool = False):
    """Wrapper to automatically set root_helper in brick calls.

    :param multipath: A boolean indicating whether the connector can
                      support multipath.
    :param enforce_multipath: If True, it raises exception when multipath=True
                              is specified but multipathd is not running.
                              If False, it falls back to multipath=False
                              when multipathd is not running.
    """

    root_helper = utils.get_root_helper()
    return connector.get_connector_properties(root_helper,
                                              CONF.my_ip,
                                              multipath,
                                              enforce_multipath)


def brick_get_connector(protocol: str,
                        driver=None,
                        use_multipath: bool = False,
                        device_scan_attempts: int = 3,
                        *args, **kwargs):
    """Wrapper to get a brick connector object.

    This automatically populates the required protocol as well
    as the root_helper needed to execute commands.
    """

    root_helper = utils.get_root_helper()
    return connector.InitiatorConnector.factory(protocol, root_helper,
                                                driver=driver,
                                                use_multipath=use_multipath,
                                                device_scan_attempts=
                                                device_scan_attempts,
                                                *args, **kwargs)


def brick_get_encryptor(connection_info: dict, *args, **kwargs):
    """Wrapper to get a brick encryptor object."""

    root_helper = utils.get_root_helper()
    km = castellan_key_manager.API(CONF)
    return encryptors.get_volume_encryptor(root_helper=root_helper,
                                           connection_info=connection_info,
                                           keymgr=km,
                                           *args, **kwargs)


def brick_attach_volume_encryptor(context: context.RequestContext,
                                  attach_info: dict,
                                  encryption: dict) -> None:
    """Attach encryption layer."""
    connection_info = attach_info['conn']
    connection_info['data']['device_path'] = attach_info['device']['path']
    encryptor = brick_get_encryptor(connection_info,
                                    **encryption)
    encryptor.attach_volume(context, **encryption)


def brick_detach_volume_encryptor(attach_info: dict, encryption: dict) -> None:
    """Detach encryption layer."""
    connection_info = attach_info['conn']
    connection_info['data']['device_path'] = attach_info['device']['path']

    encryptor = brick_get_encryptor(connection_info,
                                    **encryption)
    encryptor.detach_volume(**encryption)


# NOTE: the trace methods are included in volume_utils because
# they are currently only called by code in the volume area
# of Cinder.  These can be moved to a different file if they
# are needed elsewhere.
def trace(*dec_args, **dec_kwargs):
    """Trace calls to the decorated function.

    This decorator should always be defined as the outermost decorator so it
    is defined last. This is important so it does not interfere
    with other decorators.

    Using this decorator on a function will cause its execution to be logged at
    `DEBUG` level with arguments, return values, and exceptions.

    :returns: a function decorator
    """

    def _decorator(f):

        func_name = f.__name__

        @functools.wraps(f)
        def trace_logging_wrapper(*args, **kwargs):
            filter_function = dec_kwargs.get('filter_function')

            if len(args) > 0:
                maybe_self = args[0]
            else:
                maybe_self = kwargs.get('self', None)

            if maybe_self and hasattr(maybe_self, '__module__'):
                logger = logging.getLogger(maybe_self.__module__)
            else:
                logger = LOG

            # NOTE(ameade): Don't bother going any further if DEBUG log level
            # is not enabled for the logger.
            if not logger.isEnabledFor(py_logging.DEBUG):
                return f(*args, **kwargs)

            all_args = inspect.getcallargs(f, *args, **kwargs)

            pass_filter = filter_function is None or filter_function(all_args)

            if pass_filter:
                logger.debug('==> %(func)s: call %(all_args)r',
                             {'func': func_name,
                              'all_args': strutils.mask_password(
                                  str(all_args))})

            start_time = time.time() * 1000
            try:
                result = f(*args, **kwargs)
            except Exception as exc:
                total_time = int(round(time.time() * 1000)) - start_time
                logger.debug('<== %(func)s: exception (%(time)dms) %(exc)r',
                             {'func': func_name,
                              'time': total_time,
                              'exc': exc})
                raise
            total_time = int(round(time.time() * 1000)) - start_time

            if isinstance(result, dict):
                mask_result = strutils.mask_dict_password(result)
            elif isinstance(result, str):
                mask_result = strutils.mask_password(result)
            else:
                mask_result = result

            if pass_filter:
                logger.debug('<== %(func)s: return (%(time)dms) %(result)r',
                             {'func': func_name,
                              'time': total_time,
                              'result': mask_result})
            return result
        return trace_logging_wrapper

    if len(dec_args) == 0:
        # filter_function is passed and args does not contain f
        return _decorator
    else:
        # filter_function is not passed
        return _decorator(dec_args[0])


def trace_api(*dec_args, **dec_kwargs):
    """Decorates a function if TRACE_API is true."""

    def _decorator(f):
        @functools.wraps(f)
        def trace_api_logging_wrapper(*args, **kwargs):
            if TRACE_API:
                return trace(f, *dec_args, **dec_kwargs)(*args, **kwargs)
            return f(*args, **kwargs)
        return trace_api_logging_wrapper

    if len(dec_args) == 0:
        # filter_function is passed and args does not contain f
        return _decorator
    else:
        # filter_function is not passed
        return _decorator(dec_args[0])


def trace_method(f):
    """Decorates a function if TRACE_METHOD is true."""
    @functools.wraps(f)
    def trace_method_logging_wrapper(*args, **kwargs):
        if TRACE_METHOD:
            return trace(f)(*args, **kwargs)
        return f(*args, **kwargs)
    return trace_method_logging_wrapper


class TraceWrapperMetaclass(type):
    """Metaclass that wraps all methods of a class with trace_method.

    This metaclass will cause every function inside of the class to be
    decorated with the trace_method decorator.

    To use the metaclass you define a class like so:
    class MyClass(object, metaclass=utils.TraceWrapperMetaclass):
    """
    def __new__(meta, classname, bases, classDict):
        newClassDict = {}
        for attributeName, attribute in classDict.items():
            if isinstance(attribute, types.FunctionType):
                # replace it with a wrapped version
                attribute = functools.update_wrapper(trace_method(attribute),
                                                     attribute)
            newClassDict[attributeName] = attribute

        return type.__new__(meta, classname, bases, newClassDict)


class TraceWrapperWithABCMetaclass(abc.ABCMeta, TraceWrapperMetaclass):
    """Metaclass that wraps all methods of a class with trace."""
    pass


def setup_tracing(trace_flags):
    """Set global variables for each trace flag.

    Sets variables TRACE_METHOD and TRACE_API, which represent
    whether to log methods or api traces.

    :param trace_flags: a list of strings
    """
    global TRACE_METHOD
    global TRACE_API
    try:
        trace_flags = [flag.strip() for flag in trace_flags]
    except TypeError:  # Handle when trace_flags is None or a test mock
        trace_flags = []
    for invalid_flag in (set(trace_flags) - VALID_TRACE_FLAGS):
        LOG.warning('Invalid trace flag: %s', invalid_flag)
    TRACE_METHOD = 'method' in trace_flags
    TRACE_API = 'api' in trace_flags


def require_driver_initialized(driver):
    """Verifies if `driver` is initialized

    If the driver is not initialized, an exception will be raised.

    :params driver: The driver instance.
    :raises: `exception.DriverNotInitialized`
    """
    # we can't do anything if the driver didn't init
    if not driver.initialized:
        driver_name = driver.__class__.__name__
        LOG.error("Volume driver %s not initialized", driver_name)
        raise exception.DriverNotInitialized()
    else:
        log_unsupported_driver_warning(driver)


def log_unsupported_driver_warning(driver):
    """Annoy the log about unsupported drivers."""
    if not driver.supported:
        # Check to see if the driver is flagged as supported.
        LOG.warning("Volume driver (%(driver_name)s %(version)s) is "
                    "currently unsupported and may be removed in the "
                    "next release of OpenStack.  Use at your own risk.",
                    {'driver_name': driver.__class__.__name__,
                     'version': driver.get_version()},
                    resource={'type': 'driver',
                              'id': driver.__class__.__name__})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""RADOS Block Device iSCSI Driver"""

from distutils import version

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import netutils

from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import configuration
from cinder.volume.drivers import rbd
from cinder.volume import volume_utils

try:
    import rbd_iscsi_client
    from rbd_iscsi_client import client
    from rbd_iscsi_client import exceptions as client_exceptions
except ImportError:
    rbd_iscsi_client = None
    client = None
    client_exceptions = None

LOG = logging.getLogger(__name__)

RBD_ISCSI_OPTS = [
    cfg.StrOpt('rbd_iscsi_api_user',
               default='',
               help='The username for the rbd_target_api service'),
    cfg.StrOpt('rbd_iscsi_api_password',
               default='',
               secret=True,
               help='The username for the rbd_target_api service'),
    cfg.StrOpt('rbd_iscsi_api_url',
               default='',
               help='The url to the rbd_target_api service'),
    cfg.BoolOpt('rbd_iscsi_api_debug',
                default=False,
                help='Enable client request debugging.'),
    cfg.StrOpt('rbd_iscsi_target_iqn',
               default=None,
               help='The preconfigured target_iqn on the iscsi gateway.'),
]

CONF = cfg.CONF
CONF.register_opts(RBD_ISCSI_OPTS, group=configuration.SHARED_CONF_GROUP)


MIN_CLIENT_VERSION = "0.1.8"


@interface.volumedriver
class RBDISCSIDriver(rbd.RBDDriver):
    """Implements RADOS block device (RBD) iSCSI volume commands."""

    VERSION = '1.0.0'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Cinder_Jenkins"

    SUPPORTS_ACTIVE_ACTIVE = True

    STORAGE_PROTOCOL = 'iSCSI'
    CHAP_LENGTH = 16

    # The target IQN to use for creating all exports
    # we map all the targets for OpenStack attaches to this.
    target_iqn = None

    def __init__(self, active_backend_id=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.configuration.append_config_values(RBD_ISCSI_OPTS)

    @classmethod
    def get_driver_options(cls):
        additional_opts = cls._get_oslo_driver_opts(
            'replication_device', 'reserved_percentage',
            'max_over_subscription_ratio', 'volume_dd_blocksize',
            'driver_ssl_cert_verify', 'suppress_requests_ssl_warnings')
        return rbd.RBD_OPTS + RBD_ISCSI_OPTS + additional_opts

    def _create_client(self):
        client_version = rbd_iscsi_client.version
        if (version.StrictVersion(client_version) <
                version.StrictVersion(MIN_CLIENT_VERSION)):
            ex_msg = (_('Invalid rbd_iscsi_client version found (%(found)s). '
                        'Version %(min)s or greater required. Run "pip'
                        ' install --upgrade rbd-iscsi-client" to upgrade'
                        ' the client.')
                      % {'found': client_version,
                         'min': MIN_CLIENT_VERSION})
            LOG.error(ex_msg)
            raise exception.InvalidInput(reason=ex_msg)

        config = self.configuration
        ssl_warn = config.safe_get('suppress_requests_ssl_warnings')
        cl = client.RBDISCSIClient(
            config.safe_get('rbd_iscsi_api_user'),
            config.safe_get('rbd_iscsi_api_password'),
            config.safe_get('rbd_iscsi_api_url'),
            secure=config.safe_get('driver_ssl_cert_verify'),
            suppress_ssl_warnings=ssl_warn
        )

        return cl

    def _is_status_200(self, response):
        return (response and 'status' in response and
                response['status'] == '200')

    def do_setup(self, context):
        """Perform initialization steps that could raise exceptions."""
        super(RBDISCSIDriver, self).do_setup(context)
        if client is None:
            msg = _("You must install rbd-iscsi-client python package "
                    "before using this driver.")
            raise exception.VolumeDriverException(data=msg)

        # Make sure we have the basic settings we need to talk to the
        # iscsi api service
        config = self.configuration
        self.client = self._create_client()
        self.client.set_debug_flag(config.safe_get('rbd_iscsi_api_debug'))
        resp, body = self.client.get_api()
        if not self._is_status_200(resp):
            # failed to fetch the open api url
            raise exception.InvalidConfigurationValue(
                option='rbd_iscsi_api_url',
                value='Could not talk to the rbd-target-api')

        # The admin had to have setup a target_iqn in the iscsi gateway
        # already in order for the gateways to work properly
        self.target_iqn = self.configuration.safe_get('rbd_iscsi_target_iqn')
        LOG.info("Using target_iqn '%s'", self.target_iqn)

    def check_for_setup_error(self):
        """Return an error if prerequisites aren't met."""
        super(RBDISCSIDriver, self).check_for_setup_error()

        required_options = ['rbd_iscsi_api_user',
                            'rbd_iscsi_api_password',
                            'rbd_iscsi_api_url',
                            'rbd_iscsi_target_iqn']

        for attr in required_options:
            val = getattr(self.configuration, attr)
            if not val:
                raise exception.InvalidConfigurationValue(option=attr,
                                                          value=val)

    def _get_clients(self):
        # make sure we have
        resp, body = self.client.get_clients(self.target_iqn)
        if not self._is_status_200(resp):
            msg = _("Failed to get_clients() from rbd-target-api")
            raise exception.VolumeBackendAPIException(data=msg)
        return body

    def _get_config(self):
        resp, body = self.client.get_config()
        if not self._is_status_200(resp):
            msg = _("Failed to get_config() from rbd-target-api")
            raise exception.VolumeBackendAPIException(data=msg)
        return body

    def _get_disks(self):
        resp, disks = self.client.get_disks()
        if not self._is_status_200(resp):
            msg = _("Failed to get_disks() from rbd-target-api")
            raise exception.VolumeBackendAPIException(data=msg)
        return disks

    def create_client(self, initiator_iqn):
        """Create a client iqn on the gateway if it doesn't exist."""
        client = self._get_target_client(initiator_iqn)
        if not client:
            try:
                self.client.create_client(self.target_iqn,
                                          initiator_iqn)
            except client_exceptions.ClientException as ex:
                raise exception.VolumeBackendAPIException(
                    data=ex.get_description())

    def _get_target_client(self, initiator_iqn):
        """Get the config information for a client defined to a target."""
        config = self._get_config()
        target_config = config['targets'][self.target_iqn]
        if initiator_iqn in target_config['clients']:
            return target_config['clients'][initiator_iqn]

    def _get_auth_for_client(self, initiator_iqn):
        initiator_config = self._get_target_client(initiator_iqn)
        if initiator_config:
            auth = initiator_config['auth']
            return auth

    def _set_chap_for_client(self, initiator_iqn, username, password):
        """Save the CHAP creds in the client on the gateway."""
        # username is 8-64 chars
        # Password has to be 12-16 chars
        LOG.debug("Setting chap creds to %(user)s : %(pass)s",
                  {'user': username, 'pass': password})
        try:
            self.client.set_client_auth(self.target_iqn,
                                        initiator_iqn,
                                        username,
                                        password)
        except client_exceptions.ClientException as ex:
            raise exception.VolumeBackendAPIException(
                data=ex.get_description())

    def _get_lun(self, iscsi_config, lun_name, initiator_iqn):
        lun = None
        target_info = iscsi_config['targets'][self.target_iqn]
        luns = target_info['clients'][initiator_iqn]['luns']
        if lun_name in luns:
            lun = {'name': lun_name,
                   'id': luns[lun_name]['lun_id']}

        return lun

    def _lun_name(self, volume_name):
        """Build the iscsi gateway lun name."""
        return ("%(pool)s/%(volume_name)s" %
                {'pool': self.configuration.rbd_pool,
                 'volume_name': volume_name})

    def get_existing_disks(self):
        """Get the existing list of registered volumes on the gateway."""
        resp, disks = self.client.get_disks()
        return disks['disks']

    @volume_utils.trace
    def create_disk(self, volume_name):
        """Register the volume with the iscsi gateways.

        We have to register the volume with the iscsi gateway.
        Exporting the volume won't work unless the gateway knows
        about it.
        """
        try:
            self.client.find_disk(self.configuration.rbd_pool,
                                  volume_name)
        except client_exceptions.HTTPNotFound:
            try:
                # disk isn't known by the gateways, so lets add it.
                self.client.create_disk(self.configuration.rbd_pool,
                                        volume_name)
            except client_exceptions.ClientException as ex:
                LOG.exception("Couldn't create the disk entry to "
                              "export the volume.")
                raise exception.VolumeBackendAPIException(
                    data=ex.get_description())

    @volume_utils.trace
    def register_disk(self, target_iqn, volume_name):
        """Register the disk with the target_iqn."""
        lun_name = self._lun_name(volume_name)
        try:
            self.client.register_disk(target_iqn, lun_name)
        except client_exceptions.HTTPBadRequest as ex:
            desc = ex.get_description()
            search_str = ('is already mapped on target %(target_iqn)s' %
                          {'target_iqn': self.target_iqn})
            if desc.find(search_str):
                # The volume is already registered
                return
            else:
                LOG.error("Couldn't register the volume to the target_iqn")
                raise exception.VolumeBackendAPIException(
                    data=ex.get_description())
        except client_exceptions.ClientException as ex:
            LOG.exception("Couldn't register the volume to the target_iqn",
                          ex)
            raise exception.VolumeBackendAPIException(
                data=ex.get_description())

    @volume_utils.trace
    def unregister_disk(self, target_iqn, volume_name):
        """Unregister the volume from the gateway."""
        lun_name = self._lun_name(volume_name)
        try:
            self.client.unregister_disk(target_iqn, lun_name)
        except client_exceptions.ClientException as ex:
            LOG.exception("Couldn't unregister the volume to the target_iqn",
                          ex)
            raise exception.VolumeBackendAPIException(
                data=ex.get_description())

    @volume_utils.trace
    def export_disk(self, initiator_iqn, volume_name, iscsi_config):
        """Export a volume to an initiator."""
        lun_name = self._lun_name(volume_name)
        LOG.debug("Export lun %(lun)s", {'lun': lun_name})
        lun = self._get_lun(iscsi_config, lun_name, initiator_iqn)
        if lun:
            LOG.debug("Found existing lun export.")
            return lun

        try:
            LOG.debug("Creating new lun export for %(lun)s",
                      {'lun': lun_name})
            self.client.export_disk(self.target_iqn, initiator_iqn,
                                    self.configuration.rbd_pool,
                                    volume_name)

            resp, iscsi_config = self.client.get_config()
            return self._get_lun(iscsi_config, lun_name, initiator_iqn)
        except client_exceptions.ClientException as ex:
            raise exception.VolumeBackendAPIException(
                data=ex.get_description())

    @volume_utils.trace
    def unexport_disk(self, initiator_iqn, volume_name, iscsi_config):
        """Remove a volume from an initiator."""
        lun_name = self._lun_name(volume_name)
        LOG.debug("unexport lun %(lun)s", {'lun': lun_name})
        lun = self._get_lun(iscsi_config, lun_name, initiator_iqn)
        if not lun:
            LOG.debug("Didn't find LUN on gateway.")
            return

        try:
            LOG.debug("unexporting %(lun)s", {'lun': lun_name})
            self.client.unexport_disk(self.target_iqn, initiator_iqn,
                                      self.configuration.rbd_pool,
                                      volume_name)
        except client_exceptions.ClientException as ex:
            LOG.exception(ex)
            raise exception.VolumeBackendAPIException(
                data=ex.get_description())

    def find_client_luns(self, target_iqn, client_iqn, iscsi_config):
        """Find luns already exported to an initiator."""
        if 'targets' in iscsi_config:
            if target_iqn in iscsi_config['targets']:
                target_info = iscsi_config['targets'][target_iqn]
                if 'clients' in target_info:
                    clients = target_info['clients']
                    client = clients[client_iqn]
                    luns = client['luns']
                    return luns

    @volume_utils.trace
    def initialize_connection(self, volume, connector):
        """Export a volume to a host."""
        # create client
        initiator_iqn = connector['initiator']
        self.create_client(initiator_iqn)
        auth = self._get_auth_for_client(initiator_iqn)
        username = initiator_iqn
        if not auth['password']:
            password = volume_utils.generate_password(length=self.CHAP_LENGTH)
            self._set_chap_for_client(initiator_iqn, username, password)
        else:
            LOG.debug("using existing CHAP password")
            password = auth['password']

        # add disk for export
        iscsi_config = self._get_config()

        # First have to ensure that the disk is registered with
        # the gateways.
        self.create_disk(volume.name)
        self.register_disk(self.target_iqn, volume.name)

        iscsi_config = self._get_config()
        # Now export the disk to the initiator
        lun = self.export_disk(initiator_iqn, volume.name, iscsi_config)

        # fetch the updated config so we can get the lun id
        iscsi_config = self._get_config()
        target_info = iscsi_config['targets'][self.target_iqn]
        ips = target_info['ip_list']

        target_portal = ips[0]
        if netutils.is_valid_ipv6(target_portal):
            target_portal = "[%s]:3260" % target_portal
        else:
            target_portal = "%s:3260" % target_portal

        data = {
            'driver_volume_type': 'iscsi',
            'data': {
                'target_iqn': self.target_iqn,
                'target_portal': target_portal,
                'target_lun': lun['id'],
                'auth_method': 'CHAP',
                'auth_username': username,
                'auth_password': password,
            }
        }
        return data

    def _delete_disk(self, volume):
        """Remove the defined disk from the gateway."""

        # We only do this when we know it's not exported
        # anywhere in the gateway
        lun_name = self._lun_name(volume.name)
        config = self._get_config()

        # Now look for the disk on any exported target
        found = False
        for target_iqn in config['targets']:
            # Do we have the volume we are looking for?
            target = config['targets'][target_iqn]
            for client_iqn in target['clients'].keys():
                if lun_name in target['clients'][client_iqn]['luns']:
                    found = True

        if not found:
            # we can delete the disk definition
            LOG.info("Deleting volume definition in iscsi gateway for %s",
                     lun_name)
            self.client.delete_disk(self.configuration.rbd_pool, volume.name,
                                    preserve_image=True)

    def _terminate_connection(self, volume, initiator_iqn, target_iqn,
                              iscsi_config):
        # remove the disk from the client.
        self.unexport_disk(initiator_iqn, volume.name, iscsi_config)

        # Try to unregister the disk, since nobody is using it.
        self.unregister_disk(self.target_iqn, volume.name)

        config = self._get_config()

        # If there are no more luns exported to this initiator
        # then delete the initiator
        luns = self.find_client_luns(target_iqn, initiator_iqn, config)
        if not luns:
            LOG.debug("There aren't any more LUNs attached to %(iqn)s."
                      "So we unregister the volume and delete "
                      "the client entry",
                      {'iqn': initiator_iqn})

            try:
                self.client.delete_client(target_iqn, initiator_iqn)
            except client_exceptions.ClientException:
                LOG.warning("Tried to delete initiator %(iqn)s, but delete "
                            "failed.", {'iqns': initiator_iqn})

    def _terminate_all(self, volume, iscsi_config):
        """Find all exports of this volume for our target_iqn and detach."""
        disks = self._get_disks()
        lun_name = self._lun_name(volume.name)
        if lun_name not in disks['disks']:
            LOG.debug("Volume %s not attached anywhere.", lun_name)
            return

        for target_iqn_tmp in iscsi_config['targets']:
            if self.target_iqn != target_iqn_tmp:
                # We don't touch exports for targets
                # we aren't configured to manage.
                continue

            target = iscsi_config['targets'][self.target_iqn]
            for client_iqn in target['clients'].keys():
                if lun_name in target['clients'][client_iqn]['luns']:
                    self._terminate_connection(volume, client_iqn,
                                               self.target_iqn,
                                               iscsi_config)

        self._delete_disk(volume)

    @volume_utils.trace
    def terminate_connection(self, volume, connector, **kwargs):
        """Unexport the volume from the gateway."""
        iscsi_config = self._get_config()

        if not connector:
            # No connector was passed in, so this is a force detach
            # we need to detach the volume from the configured target_iqn.
            self._terminate_all(volume, iscsi_config)

        initiator_iqn = connector['initiator']
        self._terminate_connection(volume, initiator_iqn, self.target_iqn,
                                   iscsi_config)
        self._delete_disk(volume)
